#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from functools import lru_cache
from typing import (
    Any,
    Callable,
    ClassVar,
    Iterable,
    Literal,
    NoReturn,
    NotRequired,
    Self,
    TypeAlias,
    TypedDict,
    TypeVar,
    overload,
)

# Bokeh imports
from ..client.session import ClientSession
from ..server.session import ServerSession
from .property.bases import Property
from .property.dataspec import DataSpec
from .property.descriptors import PropertyDescriptor
from .serialization import (
    ObjectRep,
    Ref,
    Serializable,
    Serializer,
)

Setter: TypeAlias = ClientSession | ServerSession

HasPropsType = TypeVar("HasPropsType", bound=type[HasProps])

def abstract(cls: HasPropsType) -> HasPropsType: ...

def is_abstract(cls: type[HasProps]) -> bool: ...

def is_DataModel(cls: type[HasProps]) -> bool: ...

class MetaHasProps(type):

    __properties__: dict[str, Property[Any]]
    __overridden_defaults__: dict[str, Any]
    __themed_values__: dict[str, Any]

    def __new__(cls, class_name: str, bases: tuple[type, ...], class_dict: dict[str, Any]) -> HasProps: ...

    def __init__(cls, class_name: str, bases: tuple[type, ...], _) -> None: ...

    @property
    def model_class_reverse_map(cls) -> dict[str, type[HasProps]]: ...

class Local:
    ...

class Qualified:
    ...

class NonQualified:
    ...

class HasProps(Serializable, metaclass=MetaHasProps):

    _initialized: bool = ...

    _property_values: dict[str, Any] = ...
    _unstable_default_values: dict[str, Any] = ...
    _unstable_themed_values: dict[str, Any] = ...

    __view_model__: ClassVar[str]
    __view_module__: ClassVar[str]
    __qualified_model__: ClassVar[str]
    __implementation__: ClassVar[Any] # TODO: specific type
    __data_model__: ClassVar[bool]

    #def __init__(self, **properties: Any) -> None: ...

    #def __setattr__(self, name: str, value: Any) -> None: ...

    #def __getattr__(self, name: str) -> Any: ...

    def _raise_attribute_error_with_matches(self, name: str, properties: Iterable[str]) -> NoReturn: ...

    def __str__(self) -> str: ...

    __repr__ = __str__

    def equals(self, other: HasProps) -> bool: ...

    def to_serializable(self, serializer: Serializer) -> ObjectRep: ...

    def set_from_json(self, name: str, value: Any, *, setter: Setter | None = None) -> None: ...

    def update(self, **kwargs: Any) -> None: ...

    @overload
    @classmethod
    def lookup(cls, name: str, *, raises: Literal[True] = True) -> PropertyDescriptor[Any]: ...

    @overload
    @classmethod
    def lookup(cls, name: str, *, raises: Literal[False] = False) -> PropertyDescriptor[Any] | None: ...

    @classmethod
    def lookup(cls, name: str, *, raises: bool = True) -> PropertyDescriptor[Any] | None: ...

    @overload
    @classmethod
    @lru_cache(None)
    def properties(cls, *, _with_props: Literal[False] = False) -> set[str]: ...

    @overload
    @classmethod
    @lru_cache(None)
    def properties(cls, *, _with_props: Literal[True] = True) -> dict[str, Property[Any]]: ...

    @classmethod
    @lru_cache(None)
    def properties(cls, *, _with_props: bool = False) -> set[str] | dict[str, Property[Any]]: ...

    @classmethod
    @lru_cache(None)
    def descriptors(cls) -> list[PropertyDescriptor[Any]]: ...

    @classmethod
    @lru_cache(None)
    def properties_with_refs(cls) -> dict[str, Property[Any]]: ...

    @classmethod
    @lru_cache(None)
    def dataspecs(cls) -> dict[str, DataSpec]: ...

    def properties_with_values(self, *, include_defaults: bool = True, include_undefined: bool = False) -> dict[str, Any]: ...

    @classmethod
    def _overridden_defaults(cls) -> dict[str, Any]: ...

    def query_properties_with_values(self, query: Callable[[PropertyDescriptor[Any]], bool], *,
            include_defaults: bool = True, include_undefined: bool = False) -> dict[str, Any]: ...

    def themed_values(self) -> dict[str, Any] | None: ...

    def apply_theme(self, property_values: dict[str, Any]) -> None: ...

    def unapply_theme(self) -> None: ...

    def clone(self, **overrides: Any) -> Self: ...

KindRef = Any # TODO

class PropertyDef(TypedDict):
    name: str
    kind: KindRef
    default: NotRequired[Any]

class OverrideDef(TypedDict):
    name: str
    default: Any

class ModelDef(TypedDict):
    type: Literal["model"]
    name: str
    extends: NotRequired[Ref | None]
    properties: NotRequired[list[PropertyDef]]
    overrides: NotRequired[list[OverrideDef]]



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass

# Bokeh imports
from ...core.has_props import abstract
from ...core.property_aliases import IconLikeType as IconLike
from ...model import Model
from ..callbacks import Callback
from .ui_element import UIElement

@abstract
@dataclass(init=False)
class MenuItem(Model):

    checked: bool | None = ...

    icon: IconLike | None = ...

    label: str = ...

    shortcut: str | None = ...

    menu: Menu | None = ...

    tooltip: str | None = ...

    disabled: bool = ...

    action: Callback | None = ...

@dataclass
class ActionItem(MenuItem):
    ...

@dataclass
class CheckableItem(ActionItem):
    ...

@dataclass
class DividerItem(Model):
    ...

@dataclass
class Menu(UIElement):

    items: list[MenuItem | DividerItem | None] = ...

    reversed: bool = ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass
from datetime import datetime as DateTime, timedelta as TimeDelta
from typing import Sequence, TypeAlias

# Bokeh imports
from .._types import Readonly
from ..core.enums import (
    AutoType as Auto,
    PaddingUnitsType as PaddingUnits,
    StartEndType as StartEnd,
)
from ..core.has_props import abstract
from ..core.property.visual import Bounds, MinMaxBoundsType as MinMaxBounds
from ..model import Model

Value: TypeAlias = float | DateTime | TimeDelta

Interval: TypeAlias = float | TimeDelta

MinMaxInterval: TypeAlias = Auto | Bounds[float] | Bounds[TimeDelta]

L1Factor: TypeAlias = str
L2Factor: TypeAlias = tuple[str, str]
L3Factor: TypeAlias = tuple[str, str, str]

Factor: TypeAlias = L1Factor | L2Factor | L3Factor
FactorSeq: TypeAlias = Sequence[L1Factor] | Sequence[L2Factor] | Sequence[L3Factor]

@abstract
@dataclass(init=False)
class Range(Model):
    ...

@abstract
@dataclass(init=False)
class NumericalRange(Range):

    start: Value = ...

    end: Value = ...

@dataclass
class Range1d(NumericalRange):

    def __init__(self, start: Value, end: Value) -> None: ...

    reset_start: Value | None = ...

    reset_end: Value | None = ...

    bounds: MinMaxBounds | None = ...

    min_interval: Interval | None = ...

    max_interval: Interval | None = ...

@abstract
@dataclass(init=False)
class DataRange(NumericalRange):

    renderers: list[Model] | Auto | None = ...

@dataclass
class DataRange1d(DataRange):

    range_padding: Interval = ...

    range_padding_units: PaddingUnits = ...

    bounds: MinMaxBounds | None = ...

    min_interval: Interval | None = ...

    max_interval: Interval | None = ...

    flipped: bool = ...

    follow: StartEnd | None = ...

    follow_interval: Interval | None = ...

    default_span: Interval = ...

    only_visible: bool = ...

@dataclass
class FactorRange(Range):

    def __init__(self, factors: FactorSeq) -> None: ...

    factors: FactorSeq = ...

    factor_padding: float = ...

    subgroup_padding: float = ...

    group_padding: float = ...

    range_padding: float = ...

    range_padding_units: PaddingUnits = ...

    start: Readonly[float] = ...

    end: Readonly[float] = ...

    bounds: MinMaxInterval | None = ...

    min_interval: float | None = ...

    max_interval: float | None = ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass

# Bokeh imports
from ..core.has_props import abstract
from ..model import Model

@abstract
@dataclass(init=False)
class StyleSheet(Model):
    ...

@dataclass
class InlineStyleSheet(StyleSheet):

    css: str = ...

@dataclass
class ImportedStyleSheet(StyleSheet):

    url: str = ...

@dataclass
class GlobalInlineStyleSheet(InlineStyleSheet):
    ...

@dataclass
class GlobalImportedStyleSheet(ImportedStyleSheet):
    ...

@dataclass
class Styles(Model):

    align_content: str | None = ...
    align_items: str | None = ...
    align_self: str | None = ...
    alignment_baseline: str | None = ...
    all: str | None = ...
    animation: str | None = ...
    animation_delay: str | None = ...
    animation_direction: str | None = ...
    animation_duration: str | None = ...
    animation_fill_mode: str | None = ...
    animation_iteration_count: str | None = ...
    animation_name: str | None = ...
    animation_play_state: str | None = ...
    animation_timing_function: str | None = ...
    aspect_ratio: str | None = ...
    backface_visibility: str | None = ...
    background: str | None = ...
    background_attachment: str | None = ...
    background_clip: str | None = ...
    background_color: str | None = ...
    background_image: str | None = ...
    background_origin: str | None = ...
    background_position: str | None = ...
    background_position_x: str | None = ...
    background_position_y: str | None = ...
    background_repeat: str | None = ...
    background_size: str | None = ...
    baseline_shift: str | None = ...
    block_size: str | None = ...
    border: str | None = ...
    border_block_end: str | None = ...
    border_block_end_color: str | None = ...
    border_block_end_style: str | None = ...
    border_block_end_width: str | None = ...
    border_block_start: str | None = ...
    border_block_start_color: str | None = ...
    border_block_start_style: str | None = ...
    border_block_start_width: str | None = ...
    border_bottom: str | None = ...
    border_bottom_color: str | None = ...
    border_bottom_left_radius: str | None = ...
    border_bottom_right_radius: str | None = ...
    border_bottom_style: str | None = ...
    border_bottom_width: str | None = ...
    border_collapse: str | None = ...
    border_color: str | None = ...
    border_image: str | None = ...
    border_image_outset: str | None = ...
    border_image_repeat: str | None = ...
    border_image_slice: str | None = ...
    border_image_source: str | None = ...
    border_image_width: str | None = ...
    border_inline_end: str | None = ...
    border_inline_end_color: str | None = ...
    border_inline_end_style: str | None = ...
    border_inline_end_width: str | None = ...
    border_inline_start: str | None = ...
    border_inline_start_color: str | None = ...
    border_inline_start_style: str | None = ...
    border_inline_start_width: str | None = ...
    border_left: str | None = ...
    border_left_color: str | None = ...
    border_left_style: str | None = ...
    border_left_width: str | None = ...
    border_radius: str | None = ...
    border_right: str | None = ...
    border_right_color: str | None = ...
    border_right_style: str | None = ...
    border_right_width: str | None = ...
    border_spacing: str | None = ...
    border_style: str | None = ...
    border_top: str | None = ...
    border_top_color: str | None = ...
    border_top_left_radius: str | None = ...
    border_top_right_radius: str | None = ...
    border_top_style: str | None = ...
    border_top_width: str | None = ...
    border_width: str | None = ...
    bottom: str | None = ...
    box_shadow: str | None = ...
    box_sizing: str | None = ...
    break_after: str | None = ...
    break_before: str | None = ...
    break_inside: str | None = ...
    caption_side: str | None = ...
    caret_color: str | None = ...
    clear: str | None = ...
    clip: str | None = ...
    clip_path: str | None = ...
    clip_rule: str | None = ...
    color: str | None = ...
    color_interpolation: str | None = ...
    color_interpolation_filters: str | None = ...
    column_count: str | None = ...
    column_fill: str | None = ...
    column_gap: str | None = ...
    column_rule: str | None = ...
    column_rule_color: str | None = ...
    column_rule_style: str | None = ...
    column_rule_width: str | None = ...
    column_span: str | None = ...
    column_width: str | None = ...
    columns: str | None = ...
    content: str | None = ...
    counter_increment: str | None = ...
    counter_reset: str | None = ...
    cursor: str | None = ...
    direction: str | None = ...
    display: str | None = ...
    dominant_baseline: str | None = ...
    empty_cells: str | None = ...
    fill: str | None = ...
    fill_opacity: str | None = ...
    fill_rule: str | None = ...
    filter: str | None = ...
    flex: str | None = ...
    flex_basis: str | None = ...
    flex_direction: str | None = ...
    flex_flow: str | None = ...
    flex_grow: str | None = ...
    flex_shrink: str | None = ...
    flex_wrap: str | None = ...
    float: str | None = ...
    flood_color: str | None = ...
    flood_opacity: str | None = ...
    font: str | None = ...
    font_family: str | None = ...
    font_feature_settings: str | None = ...
    font_kerning: str | None = ...
    font_size: str | None = ...
    font_size_adjust: str | None = ...
    font_stretch: str | None = ...
    font_style: str | None = ...
    font_synthesis: str | None = ...
    font_variant: str | None = ...
    font_variant_caps: str | None = ...
    font_variant_east_asian: str | None = ...
    font_variant_ligatures: str | None = ...
    font_variant_numeric: str | None = ...
    font_variant_position: str | None = ...
    font_weight: str | None = ...
    gap: str | None = ...
    glyph_orientation_vertical: str | None = ...
    grid: str | None = ...
    grid_area: str | None = ...
    grid_auto_columns: str | None = ...
    grid_auto_flow: str | None = ...
    grid_auto_rows: str | None = ...
    grid_column: str | None = ...
    grid_column_end: str | None = ...
    grid_column_gap: str | None = ...
    grid_column_start: str | None = ...
    grid_gap: str | None = ...
    grid_row: str | None = ...
    grid_row_end: str | None = ...
    grid_row_gap: str | None = ...
    grid_row_start: str | None = ...
    grid_template: str | None = ...
    grid_template_areas: str | None = ...
    grid_template_columns: str | None = ...
    grid_template_rows: str | None = ...
    height: str | None = ...
    hyphens: str | None = ...
    image_orientation: str | None = ...
    image_rendering: str | None = ...
    inline_size: str | None = ...
    justify_content: str | None = ...
    justify_items: str | None = ...
    justify_self: str | None = ...
    left: str | None = ...
    letter_spacing: str | None = ...
    lighting_color: str | None = ...
    line_break: str | None = ...
    line_height: str | None = ...
    list_style: str | None = ...
    list_style_image: str | None = ...
    list_style_position: str | None = ...
    list_style_type: str | None = ...
    margin: str | None = ...
    margin_block_end: str | None = ...
    margin_block_start: str | None = ...
    margin_bottom: str | None = ...
    margin_inline_end: str | None = ...
    margin_inline_start: str | None = ...
    margin_left: str | None = ...
    margin_right: str | None = ...
    margin_top: str | None = ...
    marker: str | None = ...
    marker_end: str | None = ...
    marker_mid: str | None = ...
    marker_start: str | None = ...
    mask: str | None = ...
    mask_composite: str | None = ...
    mask_image: str | None = ...
    mask_position: str | None = ...
    mask_repeat: str | None = ...
    mask_size: str | None = ...
    mask_type: str | None = ...
    max_block_size: str | None = ...
    max_height: str | None = ...
    max_inline_size: str | None = ...
    max_width: str | None = ...
    min_block_size: str | None = ...
    min_height: str | None = ...
    min_inline_size: str | None = ...
    min_width: str | None = ...
    object_fit: str | None = ...
    object_position: str | None = ...
    opacity: str | None = ...
    order: str | None = ...
    orphans: str | None = ...
    outline: str | None = ...
    outline_color: str | None = ...
    outline_offset: str | None = ...
    outline_style: str | None = ...
    outline_width: str | None = ...
    overflow: str | None = ...
    overflow_anchor: str | None = ...
    overflow_wrap: str | None = ...
    overflow_x: str | None = ...
    overflow_y: str | None = ...
    overscroll_behavior: str | None = ...
    overscroll_behavior_block: str | None = ...
    overscroll_behavior_inline: str | None = ...
    overscroll_behavior_x: str | None = ...
    overscroll_behavior_y: str | None = ...
    padding: str | None = ...
    padding_block_end: str | None = ...
    padding_block_start: str | None = ...
    padding_bottom: str | None = ...
    padding_inline_end: str | None = ...
    padding_inline_start: str | None = ...
    padding_left: str | None = ...
    padding_right: str | None = ...
    padding_top: str | None = ...
    page_break_after: str | None = ...
    page_break_before: str | None = ...
    page_break_inside: str | None = ...
    paint_order: str | None = ...
    perspective: str | None = ...
    perspective_origin: str | None = ...
    place_content: str | None = ...
    place_items: str | None = ...
    place_self: str | None = ...
    pointer_events: str | None = ...
    position: str | None = ...
    quotes: str | None = ...
    resize: str | None = ...
    right: str | None = ...
    rotate: str | None = ...
    row_gap: str | None = ...
    ruby_align: str | None = ...
    ruby_position: str | None = ...
    scale: str | None = ...
    scroll_behavior: str | None = ...
    shape_rendering: str | None = ...
    stop_color: str | None = ...
    stop_opacity: str | None = ...
    stroke: str | None = ...
    stroke_dasharray: str | None = ...
    stroke_dashoffset: str | None = ...
    stroke_linecap: str | None = ...
    stroke_linejoin: str | None = ...
    stroke_miterlimit: str | None = ...
    stroke_opacity: str | None = ...
    stroke_width: str | None = ...
    tab_size: str | None = ...
    table_layout: str | None = ...
    text_align: str | None = ...
    text_align_last: str | None = ...
    text_anchor: str | None = ...
    text_combine_upright: str | None = ...
    text_decoration: str | None = ...
    text_decoration_color: str | None = ...
    text_decoration_line: str | None = ...
    text_decoration_style: str | None = ...
    text_emphasis: str | None = ...
    text_emphasis_color: str | None = ...
    text_emphasis_position: str | None = ...
    text_emphasis_style: str | None = ...
    text_indent: str | None = ...
    text_justify: str | None = ...
    text_orientation: str | None = ...
    text_overflow: str | None = ...
    text_rendering: str | None = ...
    text_shadow: str | None = ...
    text_transform: str | None = ...
    text_underline_position: str | None = ...
    top: str | None = ...
    touch_action: str | None = ...
    transform: str | None = ...
    transform_box: str | None = ...
    transform_origin: str | None = ...
    transform_style: str | None = ...
    transition: str | None = ...
    transition_delay: str | None = ...
    transition_duration: str | None = ...
    transition_property: str | None = ...
    transition_timing_function: str | None = ...
    translate: str | None = ...
    unicode_bidi: str | None = ...
    user_select: str | None = ...
    vertical_align: str | None = ...
    visibility: str | None = ...
    white_space: str | None = ...
    widows: str | None = ...
    width: str | None = ...
    will_change: str | None = ...
    word_break: str | None = ...
    word_spacing: str | None = ...
    word_wrap: str | None = ...
    writing_mode: str | None = ...
    z_index: str | None = ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------
''' Display a variety of visual shapes whose attributes can be associated
with data columns from ``ColumnDataSources``.

The full list of glyphs is below:

.. toctree::
   :maxdepth: 1
   :glob:

   glyphs/*

All glyphs share a minimal common interface through the base class ``Glyph``:

.. bokeh-model:: Glyph
    :module: bokeh.models.glyphs

'''

#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import annotations

import logging # isort:skip
log = logging.getLogger(__name__)

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

# Standard library imports
from typing import Any

# Bokeh imports
from ..core.enums import (
    Direction,
    HexTileOrientation,
    ImageOrigin,
    OutlineShapeName,
    Palette,
    RadiusDimension,
    StepMode,
)
from ..core.has_props import abstract
from ..core.properties import (
    AngleSpec,
    Bool,
    DataSpec,
    Dict,
    DistanceSpec,
    Either,
    Enum,
    Float,
    FloatSpec,
    Include,
    Instance,
    InstanceDefault,
    Int,
    MarkerSpec,
    NullDistanceSpec,
    NumberSpec,
    Override,
    Regex,
    Size,
    SizeSpec,
    String,
    StringSpec,
    Tuple,
    field,
    value,
)
from ..core.property_aliases import (
    Anchor,
    BorderRadius,
    Padding,
    TextAnchor,
)
from ..core.property_mixins import (
    FillProps,
    HatchProps,
    ImageProps,
    LineProps,
    ScalarFillProps,
    ScalarHatchProps,
    ScalarLineProps,
    TextProps,
)
from .callbacks import CustomJS
from .glyph import (
    ConnectedXYGlyph,
    FillGlyph,
    Glyph,
    HatchGlyph,
    LineGlyph,
    RadialGlyph,
    TextGlyph,
    XYGlyph,
)
from .mappers import ColorMapper, LinearColorMapper, StackColorMapper

#-----------------------------------------------------------------------------
# Globals and constants
#-----------------------------------------------------------------------------

__all__ = (
    'AnnularWedge',
    'Annulus',
    'Arc',
    'Bezier',
    'Block',
    'Circle',
    'ConnectedXYGlyph',
    'Ellipse',
    'Glyph',
    'HArea',
    'HAreaStep',
    'HBar',
    'HSpan',
    'HStrip',
    'HexTile',
    'Image',
    'ImageRGBA',
    'ImageStack',
    'ImageURL',
    'Line',
    'Marker',
    'MathMLGlyph',
    'MathTextGlyph',
    'MultiLine',
    'MultiPolygons',
    'Ngon',
    'Patch',
    'Patches',
    'Quad',
    'Quadratic',
    'Ray',
    'Rect',
    'Scatter',
    'Segment',
    'Step',
    'TeXGlyph',
    'Text',
    'VArea',
    'VAreaStep',
    'VBar',
    'VSpan',
    'VStrip',
    'Wedge',
    'XYGlyph',
)

#-----------------------------------------------------------------------------
# General API
#-----------------------------------------------------------------------------

@abstract
class Marker(XYGlyph, LineGlyph, FillGlyph, HatchGlyph):
    ''' Base class for glyphs that are simple markers with line and
    fill properties, located at an (x, y) location with a specified
    size.

    .. note::
        For simplicity, all markers have both line and fill properties
        declared, however some marker types (`asterisk`, `cross`, `x`)
        only draw lines. For these markers, the fill values are simply
        ignored.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    _args = ('x', 'y', 'size', 'angle')

    x = NumberSpec(default=field("x"), help="""
    The x-axis coordinates for the center of the markers.
    """)

    y = NumberSpec(default=field("y"), help="""
    The y-axis coordinates for the center of the markers.
    """)

    hit_dilation = Size(default=1.0, help="""
    The factor by which to dilate the hit radius
    which is responsible for defining the range in which a
    marker responds to interactions with the Hover and Tap
    tools.
    """)

    size = SizeSpec(default=4, help="""
    The size (diameter) values for the markers in screen space units.
    """)

    angle = AngleSpec(default=0.0, help="""
    The angles to rotate the markers.
    """)

    line_props = Include(LineProps, help="""
    The {prop} values for the markers.
    """)

    fill_props = Include(FillProps, help="""
    The {prop} values for the markers.
    """)

    hatch_props = Include(HatchProps, help="""
    The {prop} values for the markers.
    """)

@abstract
class LRTBGlyph(Glyph, LineGlyph, FillGlyph, HatchGlyph):
    """ Base class for axis-aligned rectangles. """

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    border_radius = BorderRadius(default=0, help="""
    Allows the box to have rounded corners.

    .. note::
        This property is experimental and may change at any point.
    """)

class AnnularWedge(XYGlyph, LineGlyph, FillGlyph, HatchGlyph):
    ''' Render annular wedges.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    __example__ = "examples/reference/models/AnnularWedge.py"

    _args = ('x', 'y', 'inner_radius', 'outer_radius', 'start_angle', 'end_angle', 'direction')

    x = NumberSpec(default=field("x"), help="""
    The x-coordinates of the center of the annular wedges.
    """)

    y = NumberSpec(default=field("y"), help="""
    The y-coordinates of the center of the annular wedges.
    """)

    inner_radius = DistanceSpec(default=field("inner_radius"), help="""
    The inner radii of the annular wedges.
    """)

    outer_radius = DistanceSpec(default=field("outer_radius"), help="""
    The outer radii of the annular wedges.
    """)

    start_angle = AngleSpec(default=field("start_angle"), help="""
    The angles to start the annular wedges, as measured from the horizontal.
    """)

    end_angle = AngleSpec(default=field("end_angle"), help="""
    The angles to end the annular wedges, as measured from the horizontal.
    """)

    direction = Enum(Direction, default=Direction.anticlock, help="""
    Which direction to stroke between the start and end angles.
    """)

    line_props = Include(LineProps, help="""
    The {prop} values for the annular wedges.
    """)

    fill_props = Include(FillProps, help="""
    The {prop} values for the annular wedges.
    """)

    hatch_props = Include(HatchProps, help="""
    The {prop} values for the annular wedges.
    """)

class Annulus(XYGlyph, LineGlyph, FillGlyph, HatchGlyph):
    ''' Render annuli.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    __example__ = "examples/reference/models/Annulus.py"

    _args = ('x', 'y', 'inner_radius', 'outer_radius')

    x = NumberSpec(default=field("x"), help="""
    The x-coordinates of the center of the annuli.
    """)

    y = NumberSpec(default=field("y"), help="""
    The y-coordinates of the center of the annuli.
    """)

    inner_radius = DistanceSpec(default=field("inner_radius"), help="""
    The inner radii of the annuli.
    """)

    outer_radius = DistanceSpec(default=field("outer_radius"), help="""
    The outer radii of the annuli.
    """)

    line_props = Include(LineProps, help="""
    The {prop} values for the annuli.
    """)

    fill_props = Include(FillProps, help="""
    The {prop} values for the annuli.
    """)

    hatch_props = Include(HatchProps, help="""
    The {prop} values for the annuli.
    """)

class Arc(XYGlyph, LineGlyph):
    ''' Render arcs.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    __example__ = "examples/reference/models/Arc.py"

    _args = ('x', 'y', 'radius', 'start_angle', 'end_angle', 'direction')

    x = NumberSpec(default=field("x"), help="""
    The x-coordinates of the center of the arcs.
    """)

    y = NumberSpec(default=field("y"), help="""
    The y-coordinates of the center of the arcs.
    """)

    radius = DistanceSpec(default=field("radius"), help="""
    Radius of the arc.
    """)

    start_angle = AngleSpec(default=field("start_angle"), help="""
    The angles to start the arcs, as measured from the horizontal.
    """)

    end_angle = AngleSpec(default=field("end_angle"), help="""
    The angles to end the arcs, as measured from the horizontal.
    """)

    direction = Enum(Direction, default='anticlock', help="""
    Which direction to stroke between the start and end angles.
    """)

    line_props = Include(LineProps, help="""
    The {prop} values for the arcs.
    """)

class Bezier(Glyph, LineGlyph):
    ''' Render Bezier curves.

    For more information consult the `Wikipedia article for Bezier curve`_.

    .. _Wikipedia article for Bezier curve: http://en.wikipedia.org/wiki/Bezier_curve

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    __example__ = "examples/reference/models/Bezier.py"

    _args = ("x0", "y0", "x1", "y1", "cx0", "cy0", "cx1", "cy1")

    x0 = NumberSpec(default=field("x0"), help="""
    The x-coordinates of the starting points.
    """)

    y0 = NumberSpec(default=field("y0"), help="""
    The y-coordinates of the starting points.
    """)

    x1 = NumberSpec(default=field("x1"), help="""
    The x-coordinates of the ending points.
    """)

    y1 = NumberSpec(default=field("y1"), help="""
    The y-coordinates of the ending points.
    """)

    cx0 = NumberSpec(default=field("cx0"), help="""
    The x-coordinates of first control points.
    """)

    cy0 = NumberSpec(default=field("cy0"), help="""
    The y-coordinates of first control points.
    """)

    cx1 = NumberSpec(default=field("cx1"), help="""
    The x-coordinates of second control points.
    """)

    cy1 = NumberSpec(default=field("cy1"), help="""
    The y-coordinates of second control points.
    """)

    line_props = Include(LineProps, help="""
    The {prop} values for the Bezier curves.
    """)

class Block(LRTBGlyph):
    ''' Render rectangular regions, given a lower-left corner coordinate, width, and height.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    __example__ = "examples/reference/models/Block.py"

    _args = ('x', 'y', 'width', 'height')

    x = NumberSpec(default=field("x"), help="""
    The x-coordinates of each block's lower-left corner.
    """)

    y = NumberSpec(default=field("y"), help="""
    The y-coordinates of each block's lower-left corner.
    """)

    width = DistanceSpec(default=1, help="""
    The widths of the blocks.
    """)

    height = DistanceSpec(default=1, help="""
    The heights of the blocks.
    """)

    line_props = Include(LineProps, help="""
    The {prop} values for the blocks.
    """)

    fill_props = Include(FillProps, help="""
    The {prop} values for the blocks.
    """)

    hatch_props = Include(HatchProps, help="""
    The {prop} values for the blocks.
    """)

class Circle(RadialGlyph, LineGlyph, FillGlyph, HatchGlyph):
    ''' Render circle markers. '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    __example__ = "examples/reference/models/Circle.py"

    _args = ('x', 'y', 'radius')

    x = NumberSpec(default=field("x"), help="""
    The x-coordinates of the center of the circles.
    """)

    y = NumberSpec(default=field("y"), help="""
    The y-coordinates of the center of the circles.
    """)

    radius = DistanceSpec(help="""
    The radius values for circles (in |data units|, by default).

    .. warning::
        Note that circle glyphs are always drawn as circles on the screen,
        even in cases where the data space aspect ratio is not 1-1. In all
        cases where radius values are specified, the "distance" for the radius
        is measured along the dimension specified by ``radius_dimension``. If
        the aspect ratio is very large or small, the drawn circles may appear
        much larger or smaller than expected. See :bokeh-issue:`626` for more
        information.
    """)

    radius_dimension = Enum(RadiusDimension, help="""
    What dimension to measure circle radii along.

    When the data space aspect ratio is not 1-1, then the size of the drawn
    circles depends on what direction is used to measure the "distance" of
    the radius. This property allows that direction to be controlled.

    Setting this dimension to 'max' will calculate the radius on both the x
    and y dimensions and use the maximum of the two, 'min' selects the minimum.
    """)

    hit_dilation = Size(default=1.0, help="""
    The factor by which to dilate the hit radius for hover and tap tools.
    Making this value larger makes the tools "more sensitive".
    """)

    line_props = Include(LineProps, help="""
    The {prop} values for the circles.
    """)

    fill_props = Include(FillProps, help="""
    The {prop} values for the circles.
    """)

    hatch_props = Include(HatchProps, help="""
    The {prop} values for the circles.
    """)

class Ellipse(XYGlyph, LineGlyph, FillGlyph, HatchGlyph):
    ''' Render ellipses.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    __example__ = "examples/reference/models/Ellipse.py"

    _args = ('x', 'y', 'width', 'height', 'angle')

    x = NumberSpec(default=field("x"), help="""
    The x-coordinates of the centers of the ellipses.
    """)

    y = NumberSpec(default=field("y"), help="""
    The y-coordinates of the centers of the ellipses.
    """)

    width = DistanceSpec(default=field("width"), help="""
    The widths of each ellipse.
    """)

    height = DistanceSpec(default=field("height"), help="""
    The heights of each ellipse.
    """)

    angle = AngleSpec(default=0.0, help="""
    The angle the ellipses are rotated from horizontal. [rad]
    """)

    line_props = Include(LineProps, help="""
    The {prop} values for the ellipses.
    """)

    fill_props = Include(FillProps, help="""
    The {prop} values for the ellipses.
    """)

    hatch_props = Include(HatchProps, help="""
    The {prop} values for the ellipses.
    """)

class HArea(Glyph, FillGlyph, HatchGlyph):
    ''' Render a horizontally directed area between two equal length sequences
    of x-coordinates with the same y-coordinates.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    __example__ = "examples/reference/models/HArea.py"

    _args = ('x1', 'x2', 'y')

    x1 = NumberSpec(default=field("x1"), help="""
    The x-coordinates for the points of one side of the area.
    """)

    x2 = NumberSpec(default=field("x2"), help="""
    The x-coordinates for the points of the other side of the area.
    """)

    y = NumberSpec(default=field("y"), help="""
    The y-coordinates for the points of the area.
    """)

    fill_props = Include(ScalarFillProps, help="""
    The {prop} values for the horizontal directed area.
    """)

    hatch_props = Include(HatchProps, help="""
    The {prop} values for the horizontal directed area.
    """)

class HAreaStep(Glyph, FillGlyph, HatchGlyph):
    ''' Render a horizontally directed area between two equal length sequences
    of x-coordinates with the same y-coordinates using step lines.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    __example__ = "examples/reference/models/HAreaStep.py"

    _args = ('x1', 'x2', 'y')

    x1 = NumberSpec(default=field("x1"), help="""
    The x-coordinates for the points of one side of the area.
    """)

    x2 = NumberSpec(default=field("x2"), help="""
    The x-coordinates for the points of the other side of the area.
    """)

    y = NumberSpec(default=field("y"), help="""
    The y-coordinates for the points of the area.
    """)

    step_mode = Enum(StepMode, default="before", help="""
    Where the step "level" should be drawn in relation to the x and y
    coordinates. The parameter can assume one of three values:

    * ``before``: (default) Draw step levels before each y-coordinate (no step before the first point)
    * ``after``:  Draw step levels after each y-coordinate (no step after the last point)
    * ``center``: Draw step levels centered on each y-coordinate
    """)

    fill_props = Include(ScalarFillProps, help="""
    The {prop} values for the horizontal directed area.
    """)

    hatch_props = Include(HatchProps, help="""
    The {prop} values for the horizontal directed area.
    """)

class HBar(LRTBGlyph):
    ''' Render horizontal bars, given a center coordinate, ``height`` and
    (``left``, ``right``) coordinates.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    __example__ = "examples/reference/models/HBar.py"

    _args = ('y', 'height', 'right', 'left')

    y = NumberSpec(default=field("y"), help="""
    The y-coordinates of the centers of the horizontal bars.
    """)

    height = DistanceSpec(default=1, help="""
    The heights of the vertical bars.
    """)

    left = NumberSpec(default=0, help="""
    The x-coordinates of the left edges.
    """)

    right = NumberSpec(default=field("right"), help="""
    The x-coordinates of the right edges.
    """)

    line_props = Include(LineProps, help="""
    The {prop} values for the horizontal bars.
    """)

    fill_props = Include(FillProps, help="""
    The {prop} values for the horizontal bars.
    """)

    hatch_props = Include(HatchProps, help="""
    The {prop} values for the horizontal bars.
    """)

class HexTile(Glyph, LineGlyph, FillGlyph, HatchGlyph):
    ''' Render horizontal tiles on a regular hexagonal grid.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    __example__ = "examples/reference/models/HexTile.py"

    _args = ('q', 'r')

    size = Float(1.0, help="""
    The radius (in |data units|) of the hex tiling.

    The radius is always measured along the cartesian y-axis for "pointy_top"
    orientation, and along the cartesian x-axis for "flat_top" orientation. If
    the aspect ratio of the underlying cartesian system is not 1-1, then the
    tiles may be "squished" in one direction. To ensure that the tiles are
    always regular hexagons, consider setting the ``match_aspect`` property of
    the plot to True.
    """)

    aspect_scale = Float(default=1.0, help="""
    Match a plot's aspect ratio scaling.

    Use this parameter to match the aspect ratio scaling of a plot when using
    :class:`~bokeh.models.Plot.aspect_scale` with a value other than ``1.0``.

    """)

    r = NumberSpec(default=field("r"), help="""
    The "row" axial coordinates of the tile centers.
    """)

    q = NumberSpec(default=field("q"), help="""
    The "column" axial coordinates of the tile centers.
    """)

    scale = NumberSpec(1.0, help="""
    A scale factor for individual tiles.
    """)

    orientation = Enum(HexTileOrientation, default="pointytop", help="""
    The orientation of the hex tiles.

    Use ``"pointytop"`` to orient the tile so that a pointed corner is at the top. Use
    ``"flattop"`` to orient the tile so that a flat side is at the top.
    """)

    line_props = Include(LineProps, help="""
    The {prop} values for the hex tiles.
    """)

    line_color = Override(default=None)

    fill_props = Include(FillProps, help="""
    The {prop} values for the hex tiles.
    """)

    hatch_props = Include(HatchProps, help="""
    The {prop} values for the hex tiles.
    """)

@abstract
class ImageBase(XYGlyph):

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    x = NumberSpec(default=field("x"), help="""
    The x-coordinates to locate the image anchors.
    """)

    y = NumberSpec(default=field("y"), help="""
    The y-coordinates to locate the image anchors.
    """)

    dw = DistanceSpec(default=field("dw"), help="""
    The widths of the plot regions that the images will occupy.

    .. note::
        This is not the number of pixels that an image is wide.
        That number is fixed by the image itself.
    """)

    dh = DistanceSpec(default=field("dh"), help="""
    The height of the plot region that the image will occupy.

    .. note::
        This is not the number of pixels that an image is tall.
        That number is fixed by the image itself.
    """)

    image_props = Include(ImageProps, help="""
    The {prop} values for the images.
    """)

    dilate = Bool(False, help="""
    Whether to always round fractional pixel locations in such a way
    as to make the images bigger.

    This setting may be useful if pixel rounding errors are causing
    images to have a gap between them, when they should appear flush.
    """)

    origin = Enum(ImageOrigin, default="bottom_left", help="""
    Defines the coordinate space of an image.
    """)

    anchor = Anchor(default="bottom_left", help="""
    Position of the image should be anchored at the `x`, `y` coordinates.
    """)

class Image(ImageBase):
    ''' Render images given as scalar data together with a color mapper.

    In addition to the defined model properties, ``Image`` also can accept
    a keyword argument ``palette`` in place of an explicit ``color_mapper``.
    The value should be a list of colors, or the name of one of the built-in
    palettes in ``bokeh.palettes``. This palette will be used to automatically
    construct a ``ColorMapper`` model for the ``color_mapper`` property.

    If both ``palette`` and ``color_mapper`` are passed, a ``ValueError``
    exception will be raised. If neither is passed, then the ``Greys9``
    palette will be used as a default.

    '''

    def __init__(self, *args, **kwargs) -> None:
        if 'palette' in kwargs and 'color_mapper' in kwargs:
            raise ValueError("only one of 'palette' and 'color_mapper' may be specified")
        elif 'color_mapper' not in kwargs:
            # Use a palette (given or default)
            palette = kwargs.pop('palette', 'Greys9')
            mapper = LinearColorMapper(palette)
            kwargs['color_mapper'] = mapper

        super().__init__(*args, **kwargs)

    _args = ('image', 'x', 'y', 'dw', 'dh', 'dilate')

    # a hook to specify any additional kwargs handled by an initializer
    _extra_kws = {
        'palette': (
            'str or list[color value]',
            'a palette to construct a value for the color mapper property from',
        ),
    }

    image = NumberSpec(default=field("image"), help="""
    The arrays of scalar data for the images to be colormapped.
    """)

    color_mapper = Instance(ColorMapper, default=InstanceDefault(LinearColorMapper, palette="Greys9"), help="""
    A ``ColorMapper`` to use to map the scalar data from ``image``
    into RGBA values for display.

    The name of a palette from ``bokeh.palettes`` may also be set, in which
    case a ``LinearColorMapper`` configured with the named palette will be used.

    .. note::
        The color mapping step happens on the client.
    """).accepts(Enum(Palette), lambda pal: LinearColorMapper(palette=pal))

class ImageRGBA(ImageBase):
    ''' Render images given as RGBA data.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    _args = ('image', 'x', 'y', 'dw', 'dh', 'dilate')

    image = NumberSpec(default=field("image"), help="""
    The arrays of RGBA data for the images.
    """)

class ImageStack(ImageBase):
    ''' Render images given as 3D stacked arrays by flattening each stack into
    an RGBA image using a ``StackColorMapper``.

    The 3D arrays have shape (ny, nx, nstack) where ``nstack`` is the number of
    stacks. The ``color_mapper`` produces an RGBA value for each of the
    (ny, nx) pixels by combining array values in the ``nstack`` direction.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    _args = ('image', 'x', 'y', 'dw', 'dh', 'dilate')

    image = NumberSpec(default=field("image"), help="""
    The 3D arrays of data for the images.
    """)

    color_mapper = Instance(StackColorMapper, help="""
    ``ScalarColorMapper`` used to map the scalar data from ``image``
    into RGBA values for display.

    .. note::
        The color mapping step happens on the client.
    """)

class ImageURL(XYGlyph):
    ''' Render images loaded from given URLs.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    __example__ = "examples/reference/models/ImageURL.py"

    _args = ('url', 'x', 'y', 'w', 'h', 'angle', 'dilate')

    url = StringSpec(default=field("url"), help="""
    The URLs to retrieve images from.

    .. note::
        The actual retrieving and loading of the images happens on
        the client.
    """)

    x = NumberSpec(default=field("x"), help="""
    The x-coordinates to locate the image anchors.
    """)

    y = NumberSpec(default=field("y"), help="""
    The y-coordinates to locate the image anchors.
    """)

    w = NullDistanceSpec(help="""
    The width of the plot region that the image will occupy in data space.

    The default value is ``None``, in which case the image will be displayed
    at its actual image size (regardless of the units specified here).
    """)

    h = NullDistanceSpec(help="""
    The height of the plot region that the image will occupy in data space.

    The default value is ``None``, in which case the image will be displayed
    at its actual image size (regardless of the units specified here).
    """)

    angle = AngleSpec(default=0, help="""
    The angles to rotate the images, as measured from the horizontal.
    """)

    global_alpha = NumberSpec(1.0, help="""
    An overall opacity that each image is rendered with (in addition
    to any inherent alpha values in the image itself).
    """)

    dilate = Bool(False, help="""
    Whether to always round fractional pixel locations in such a way
    as to make the images bigger.

    This setting may be useful if pixel rounding errors are causing
    images to have a gap between them, when they should appear flush.
    """)

    anchor = Anchor(default="top_left", help="""
    Position of the image should be anchored at the `x`, `y` coordinates.
    """)

    retry_attempts = Int(0, help="""
    Number of attempts to retry loading the images from the specified URL.
    Default is zero.
    """)

    retry_timeout = Int(0, help="""
    Timeout (in ms) between retry attempts to load the image from the
    specified URL. Default is zero ms.
    """)

class Line(ConnectedXYGlyph, LineGlyph):
    ''' Render a single line.

    The ``Line`` glyph is different from most other glyphs in that the vector
    of values only produces one glyph on the Plot.

    .. note::
        Due to limitations in the underlying HTML canvas, it is possible that a
        line is not drawn when one or more of its coordinates is very far outside
        the viewport. This behavior is different for different browsers. See
        :bokeh-issue:`11498` for more information.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    _args = ('x', 'y')

    __example__ = "examples/reference/models/Line.py"

    x = NumberSpec(default=field("x"), help="""
    The x-coordinates for the points of the line.
    """)

    y = NumberSpec(default=field("y"), help="""
    The y-coordinates for the points of the line.
    """)

    line_props = Include(ScalarLineProps, help="""
    The {prop} values for the line.
    """)

class MultiLine(Glyph, LineGlyph):
    ''' Render several lines.

    The data for the ``MultiLine`` glyph is different in that the vector of
    values is not a vector of scalars. Rather, it is a "list of lists".

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    __example__ = "examples/reference/models/MultiLine.py"

    _args = ('xs', 'ys')

    xs = NumberSpec(default=field("xs"), help="""
    The x-coordinates for all the lines, given as a "list of lists".
    """)

    ys = NumberSpec(default=field("ys"), help="""
    The y-coordinates for all the lines, given as a "list of lists".
    """)

    line_props = Include(LineProps, help="""
    The {prop} values for the lines.
    """)

class MultiPolygons(Glyph, LineGlyph, FillGlyph, HatchGlyph):
    ''' Render several MultiPolygon.

    Modeled on geoJSON - the data for the ``MultiPolygons`` glyph is
    different in that the vector of values is not a vector of scalars.
    Rather, it is a "list of lists of lists of lists".

    During box selection only multi-polygons entirely contained in the
    selection box will be included.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    __example__ = "examples/reference/models/MultiPolygons.py"

    _args = ('xs', 'ys')

    xs = NumberSpec(default=field("xs"), help="""
    The x-coordinates for all the patches, given as a nested list.

    .. note::
        Each item in ``MultiPolygons`` represents one MultiPolygon and each
        MultiPolygon is comprised of ``n`` Polygons. Each Polygon is made of
        one exterior ring optionally followed by ``m`` interior rings (holes).
    """)

    ys = NumberSpec(default=field("ys"), help="""
    The y-coordinates for all the patches, given as a "list of lists".

    .. note::
        Each item in ``MultiPolygons`` represents one MultiPolygon and each
        MultiPolygon is comprised of ``n`` Polygons. Each Polygon is made of
        one exterior ring optionally followed by ``m`` interior rings (holes).
    """)

    line_props = Include(LineProps, help="""
    The {prop} values for the multipolygons.
    """)

    fill_props = Include(FillProps, help="""
    The {prop} values for the multipolygons.
    """)

    hatch_props = Include(HatchProps, help="""
    The {prop} values for the multipolygons.
    """)

class Ngon(RadialGlyph):
    ''' Render regular n-sided polygons.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    __example__ = "examples/reference/models/Ngon.py"

    _args = ('x', 'y', 'radius')

    x = NumberSpec(default=field("x"), help="""
    The x-coordinates of the center of the n-gons.
    """)

    y = NumberSpec(default=field("y"), help="""
    The y-coordinates of the center of the n-gons.
    """)

    radius = DistanceSpec(help="""
    The radius values for n-gons (in |data units|, by default). The radius is
    measured from the center to the vertices of the n-gons.
    """)

    angle = AngleSpec(default=0, help="""
    The angles in radians to rotate the n-gons. When the value is zero, a vertex
    is drawn directly above the center coordinate.
    """)

    n = NumberSpec(default=field("n"), help="""
    The number of sides of the n-gons. Values less than three will result in
    no glyph instance being drawn.
    """)

    radius_dimension = Enum(RadiusDimension, help="""
    What dimension to measure n-gons radii along.

    When the data space aspect ratio is not 1-1, then the size of the drawn
    n-gons depends on what direction is used to measure the "distance" of
    the radius. This property allows that direction to be controlled.

    Setting this dimension to 'max' will calculate the radius on both the x
    and y dimensions and use the maximum of the two, 'min' selects the minimum.
    """)

    line_props = Include(LineProps, help="""
    The {prop} values for the n-gons.
    """)

    fill_props = Include(FillProps, help="""
    The {prop} values for the n-gons.
    """)

    hatch_props = Include(HatchProps, help="""
    The {prop} values for the n-gons.
    """)

class Patch(ConnectedXYGlyph, LineGlyph, FillGlyph, HatchGlyph):
    ''' Render a single patch.

    The ``Patch`` glyph is different from most other glyphs in that the vector
    of values only produces one glyph on the Plot.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    __example__ = "examples/reference/models/Patch.py"

    _args = ('x', 'y')

    x = NumberSpec(default=field("x"), help="""
    The x-coordinates for the points of the patch.

    .. note::
        A patch may comprise multiple polygons. In this case the
        x-coordinates for each polygon should be separated by NaN
        values in the sequence.
    """)

    y = NumberSpec(default=field("y"), help="""
    The y-coordinates for the points of the patch.

    .. note::
        A patch may comprise multiple polygons. In this case the
        y-coordinates for each polygon should be separated by NaN
        values in the sequence.
    """)

    line_props = Include(ScalarLineProps, help="""
    The {prop} values for the patch.
    """)

    fill_props = Include(ScalarFillProps, help="""
    The {prop} values for the patch.
    """)

    hatch_props = Include(ScalarHatchProps, help="""
    The {prop} values for the patch.
    """)

class Patches(Glyph, LineGlyph, FillGlyph, HatchGlyph):
    ''' Render several patches.

    The data for the ``Patches`` glyph is different in that the vector of
    values is not a vector of scalars. Rather, it is a "list of lists".

    During box selection only patches entirely contained in the
    selection box will be included.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    __example__ = "examples/reference/models/Patches.py"

    _args = ('xs', 'ys')

    xs = NumberSpec(default=field("xs"), help="""
    The x-coordinates for all the patches, given as a "list of lists".

    .. note::
        Individual patches may comprise multiple polygons. In this case
        the x-coordinates for each polygon should be separated by NaN
        values in the sublists.
    """)

    ys = NumberSpec(default=field("ys"), help="""
    The y-coordinates for all the patches, given as a "list of lists".

    .. note::
        Individual patches may comprise multiple polygons. In this case
        the y-coordinates for each polygon should be separated by NaN
        values in the sublists.
    """)

    line_props = Include(LineProps, help="""
    The {prop} values for the patches.
    """)

    fill_props = Include(FillProps, help="""
    The {prop} values for the patches.
    """)

    hatch_props = Include(HatchProps, help="""
    The {prop} values for the patches.
    """)

class Quad(LRTBGlyph):
    ''' Render axis-aligned quads.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    __example__ = "examples/reference/models/Quad.py"

    _args = ('left', 'right', 'top', 'bottom')

    left = NumberSpec(default=field("left"), help="""
    The x-coordinates of the left edges.
    """)

    right = NumberSpec(default=field("right"), help="""
    The x-coordinates of the right edges.
    """)

    bottom = NumberSpec(default=field("bottom"), help="""
    The y-coordinates of the bottom edges.
    """)

    top = NumberSpec(default=field("top"), help="""
    The y-coordinates of the top edges.
    """)

    line_props = Include(LineProps, help="""
    The {prop} values for the quads.
    """)

    fill_props = Include(FillProps, help="""
    The {prop} values for the quads.
    """)

    hatch_props = Include(HatchProps, help="""
    The {prop} values for the quads.
    """)

class Quadratic(Glyph, LineGlyph):
    ''' Render parabolas.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    __example__ = "examples/reference/models/Quadratic.py"

    _args = ("x0", "y0", "x1", "y1", "cx", "cy")

    x0 = NumberSpec(default=field("x0"), help="""
    The x-coordinates of the starting points.
    """)

    y0 = NumberSpec(default=field("y0"), help="""
    The y-coordinates of the starting points.
    """)

    x1 = NumberSpec(default=field("x1"), help="""
    The x-coordinates of the ending points.
    """)

    y1 = NumberSpec(default=field("y1"), help="""
    The y-coordinates of the ending points.
    """)

    cx = NumberSpec(default=field("cx"), help="""
    The x-coordinates of the control points.
    """)

    cy = NumberSpec(default=field("cy"), help="""
    The y-coordinates of the control points.
    """)

    line_props = Include(LineProps, help="""
    The {prop} values for the parabolas.
    """)

class Ray(XYGlyph, LineGlyph):
    ''' Render rays.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    __example__ = "examples/reference/models/Ray.py"

    _args = ('x', 'y', 'length', 'angle')

    x = NumberSpec(default=field("x"), help="""
    The x-coordinates to start the rays.
    """)

    y = NumberSpec(default=field("y"), help="""
    The y-coordinates to start the rays.
    """)

    angle = AngleSpec(default=0, help="""
    The angles in radians to extend the rays, as measured from the horizontal.
    """)

    length = DistanceSpec(default=0, help="""
    The length to extend the ray. Note that this ``length`` defaults
    to |data units| (measured in the x-direction).
    """)

    line_props = Include(LineProps, help="""
    The {prop} values for the rays.
    """)

class Rect(XYGlyph, LineGlyph, FillGlyph, HatchGlyph):
    ''' Render rectangles, characterised by center position (x and y), width,
    height, and angle of rotation.

    .. warning::
        ``Rect`` glyphs are not well defined on logarithmic scales. Use
        :class:`~bokeh.models.Block` or :class:`~bokeh.models.Quad` glyphs
        instead.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    __example__ = "examples/reference/models/Rect.py"

    _args = ('x', 'y', 'width', 'height', 'angle', 'dilate')

    x = NumberSpec(default=field("x"), help="""
    The x-coordinates of the centers of the rectangles.
    """)

    y = NumberSpec(default=field("y"), help="""
    The y-coordinates of the centers of the rectangles.
    """)

    width = DistanceSpec(default=field("width"), help="""
    The overall widths of the rectangles.
    """)

    height = DistanceSpec(default=field("height"), help="""
    The overall heights of the rectangles.
    """)

    angle = AngleSpec(default=0.0, help="""
    The angles to rotate the rectangles, as measured from the horizontal.
    """)

    border_radius = BorderRadius(default=0, help="""
    Allows the box to have rounded corners.

    .. note::
        This property is experimental and may change at any point.
    """)

    dilate = Bool(False, help="""
    Whether to always round fractional pixel locations in such a way
    as to make the rectangles bigger.

    This setting may be useful if pixel rounding errors are causing
    rectangles to have a gap between them, when they should appear
    flush.
    """)

    line_props = Include(LineProps, help="""
    The {prop} values for the rectangles.
    """)

    fill_props = Include(FillProps, help="""
    The {prop} values for the rectangles.
    """)

    hatch_props = Include(HatchProps, help="""
    The {prop} values for the rectangles.
    """)

class Scatter(Marker):
    ''' Render scatter markers selected from a predefined list of designs.

    Use ``Scatter`` to draw any of Bokeh's built-in marker types:
    ``asterisk``, ``circle``, ``circle_cross``, ``circle_dot``, ``circle_x``,
    ``circle_y``, ``cross``, ``dash``, ``diamond``, ``diamond_cross``,
    ``diamond_dot``, ``dot``, ``hex``, ``hex_dot``, ``inverted_triangle``,
    ``plus``, ``square``, ``square_cross``, ``square_dot``, ``square_pin``,
    ``square_x``, ``star``, ``star_dot``, ``triangle``, ``triangle_dot``,
    ``triangle_pin``, ``x``, or ``y``. This collection is available in
    :class:`~bokeh.core.enums.MarkerType`.

    Bokeh's built-in markers consist of a set of base markers, most of which can
    be combined with different kinds of additional visual features:

    .. bokeh-plot:: __REPO__/examples/basic/scatters/markertypes.py
        :source-position: none

    You can select marker types in two ways:

    * To draw the **same marker for all values**, use the ``marker`` attribute
      to specify the name of a specific marker. For example:

      .. code-block:: python

          glyph = Scatter(x="x", y="y", size="sizes", marker="square")
          plot.add_glyph(source, glyph)

      This will render square markers for all points.

    * Alternatively, to use **marker types specified in a data source column**,
      assign the column name to the ``marker`` attribute. For example:

      .. code-block:: python

          # source.data['markers'] = ["circle", "square", "circle", ... ]

          glyph = Scatter(x="x", y="y", size="sizes", marker="markers")
          plot.add_glyph(source, glyph)

    It is also possible to define a custom marker. See :attr:`bokeh.models.Scatter.defs`.

    .. note::
        When you draw ``circle`` markers with ``Scatter``, you can only assign a
        size in |screen units| (by passing a number of pixels to the ``size``
        property). In case you want to define the radius of circles in
        |data units|, use the :class:`~bokeh.models.glyphs.Circle` glyph instead
        of the ``Scatter`` glyph.

    .. note::
        ``Scatter`` markers with multiple marker types may be drawn in a
        different order when using the WebGL output backend. This is an explicit
        trade-off made in the interests of performance.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    __example__ = "examples/reference/models/Scatter.py"

    _args = ('x', 'y', 'size', 'angle', 'marker')

    marker = MarkerSpec(default="circle", help="""
    Which marker to render. This can be the name of any built in marker,
    e.g. "circle", or a reference to a data column containing such names.
    """)

    defs = Dict(Regex("^@.*$"), Instance(CustomJS))(default={}, help="""
    A collection of custom marker definitions.

    There are two ways to define a custom marker:

    * construct and return an instance of ``Path2D``:

    .. code:: python

        CustomJS(code='''
            export default (args, obj, {ctx, i, r, visuals}) => {
                const path = new Path2D()
                path.arc(0, 0, r, 0, 2*Math.PI, false)
                return path
            }
        ''')

    * paint directly to an instance of ``Context2d``:

    .. code:: python

        CustomJS(code='''
            export default (args, obj, {ctx, i, r, visuals}) => {
                ctx.arc(0, 0, r, 0, 2*Math.PI, false)
                visuals.fill.apply(ctx, i)
                visuals.hatch.apply(ctx, i)
                visuals.line.apply(ctx, i)
            }
        ''')

    .. note::

        Custom marker's names must start with `"@"` prefix, e.g. `"@my_marker"`.

    .. note::

        Custom markers are only supported with ``"canvas"`` and ``"svg"`` backends.

    """)

class Segment(Glyph, LineGlyph):
    ''' Render segments.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    __example__ = "examples/reference/models/Segment.py"

    _args = ('x0', 'y0', 'x1', 'y1')

    x0 = NumberSpec(default=field("x0"), help="""
    The x-coordinates of the starting points.
    """)

    y0 = NumberSpec(default=field("y0"), help="""
    The y-coordinates of the starting points.
    """)

    x1 = NumberSpec(default=field("x1"), help="""
    The x-coordinates of the ending points.
    """)

    y1 = NumberSpec(default=field("y1"), help="""
    The y-coordinates of the ending points.
    """)

    line_props = Include(LineProps, help="""
    The {prop} values for the segments.
    """)

class Step(XYGlyph, LineGlyph):
    ''' Render step lines.

    Step levels can be draw before, after, or centered on each point, according
    to the value of the ``mode`` property.

    The x-coordinates are assumed to be (and must be) sorted in ascending order
    for steps to be properly rendered.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    __example__ = "examples/reference/models/Step.py"

    _args = ('x', 'y')

    x = NumberSpec(default=field("x"), help="""
    The x-coordinates for the steps.
    """)

    y = NumberSpec(default=field("y"), help="""
    The y-coordinates for the steps.
    """)

    line_props = Include(ScalarLineProps, help="""
    The {prop} values for the steps.
    """)

    mode = Enum(StepMode, default="before", help="""
    Where the step "level" should be drawn in relation to the x and y
    coordinates. The parameter can assume one of three values:

    * ``before``: (default) Draw step levels before each x-coordinate (no step before the first point)
    * ``after``:  Draw step levels after each x-coordinate (no step after the last point)
    * ``center``: Draw step levels centered on each x-coordinate
    """)

class Text(XYGlyph, TextGlyph):
    ''' Render text.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    __example__ = "examples/reference/models/Text.py"

    _args = ('x', 'y', 'text', 'angle', 'x_offset', 'y_offset')

    x = NumberSpec(default=field("x"), help="""
    The x-coordinates to locate the text anchors.
    """)

    y = NumberSpec(default=field("y"), help="""
    The y-coordinates to locate the text anchors.
    """)

    text = StringSpec(default=field("text"), help="""
    The text values to render.
    """)

    angle = AngleSpec(default=0, help="""
    The angles to rotate the text, as measured from the horizontal.
    """)

    x_offset = FloatSpec(default=0, help="""
    Offset values in pixels to apply to the x-coordinates.

    This is useful, for instance, if it is desired to "float" text a fixed
    distance in |screen units| from a given data position.
    """)

    y_offset = FloatSpec(default=0, help="""
    Offset values in pixels to apply to the y-coordinates.

    This is useful, for instance, if it is desired to "float" text a fixed
    distance in |screen units| from a given data position.
    """)

    anchor = DataSpec(TextAnchor, default=value("auto"), help="""
    Position within the bounding box of this glyph to which ``x`` and ``y``
    coordinates are anchored to. This can be a named anchor point like
    ``top_left`` or ``center``, or a percentage from from left to right
    and top to bottom, or a combination of those, independently in width
    and height. If set to ``auto``, then anchor point will be determined
    from text ``align`` and ``baseline``.

    .. note::
        This property is experimental and may change at any point.
    """)

    padding = Padding(default=0, help="""
    Extra space between the text of a glyphs and its bounding box (border).

    .. note::
        This property is experimental and may change at any point.
    """)

    border_radius = BorderRadius(default=0, help="""
    Allows the box to have rounded corners. For the best results, it
    should be used in combination with ``padding``.

    .. note::
        This property is experimental and may change at any point.
    """)

    outline_shape = DataSpec(Enum(OutlineShapeName), default="box", help="""
    Specify the shape of the outline for the text box.

    The default outline is of a text box is its bounding box (or rectangle).
    This can be changed to a selection of pre-defined shapes, like circle,
    ellipse, diamond, parallelogram, etc. Those shapes are circumscribed onto
    the bounding box, so that the contents of a box fit inside those shapes.

    This property is in effect only when either border line, background fill
    and/or background hatch properties are set. The user can choose ``"none"``
    to avoid drawing any shape, even if border or background visuals are set.

    .. note::
        This property is experimental and may change at any point.

    .. note::
        Currently hit testing only uses the bounding box of text contents
        of the glyph, which is equivalent to using box/rectangle shape.
    """)

    text_props = Include(TextProps, help="""
    The {prop} values for the text.
    """)

    background_fill_props = Include(FillProps, prefix="background", help="""
    The {prop} values for the text bounding box.
    """)

    background_hatch_props = Include(HatchProps, prefix="background", help="""
    The {prop} values for the text bounding box.
    """)

    border_line_props = Include(LineProps, prefix="border", help="""
    The {prop} values for the text bounding box.
    """)

    background_fill_color = Override(default=None)

    background_hatch_color = Override(default=None)

    border_line_color = Override(default=None)

@abstract
class MathTextGlyph(Text):
    """ Base class for math text glyphs.

    """

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

class MathMLGlyph(MathTextGlyph):
    """ Render mathematical content using `MathML <https://www.w3.org/Math/>`_
    notation.

    See :ref:`ug_styling_mathtext` in the |user guide| for more information.

    """

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

class TeXGlyph(MathTextGlyph):
    """
    Render mathematical content using `LaTeX <https://www.latex-project.org/>`_
    notation.

    See :ref:`ug_styling_mathtext` in the |user guide| for more information.

    .. note::
        Bokeh uses `MathJax <https://www.mathjax.org>`_ to render text
        containing mathematical notation.

        MathJax only supports math-mode macros (no text-mode macros). You
        can see more about differences between standard TeX/LaTeX and MathJax
        here: https://docs.mathjax.org/en/latest/input/tex/differences.html

    """

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    macros = Dict(String, Either(String, Tuple(String, Int)), help="""
    User defined TeX macros.

    This is a mapping from control sequence names (without leading backslash) to
    either replacement strings or tuples of a replacement string and a number
    of arguments.

    Example:

    .. code-block:: python

        TeX(text=r"\\R \\rightarrow \\R^2", macros={"RR": r"{\\bf R}"})

    """)

    display = Enum("inline", "block", "auto", default="auto", help="""
    Defines how the text is interpreted and what TeX display mode to use.

    The following values are allowed:

    * ``"auto"`` (the default)
      The text is parsed, requiring TeX delimiters to enclose math content,
      e.g. ``"$$x^2$$"`` or ``r"\\[\\frac{x}{y}\\]"``. This allows mixed
      math text and regular text content. TeX display mode is inferred by
      the parser.
    * ``"block"``
      The text is taken verbatim and TeX's block mode is used.
    * ``"inline"``
      The text is taken verbatim and TeX's inline mode is used.
    """)

class VArea(Glyph, FillGlyph, HatchGlyph):
    ''' Render a vertically directed area between two equal length sequences
    of y-coordinates with the same x-coordinates.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    __example__ = "examples/reference/models/VArea.py"

    _args = ('x', 'y1', 'y2')

    x = NumberSpec(default=field("x"), help="""
    The x-coordinates for the points of the area.
    """)

    y1 = NumberSpec(default=field("y1"), help="""
    The y-coordinates for the points of one side of the area.
    """)

    y2 = NumberSpec(default=field("y2"), help="""
    The y-coordinates for the points of the other side of the area.
    """)

    fill_props = Include(ScalarFillProps, help="""
    The {prop} values for the vertical directed area.
    """)

    hatch_props = Include(HatchProps, help="""
    The {prop} values for the vertical directed area.
    """)

class VAreaStep(Glyph, FillGlyph, HatchGlyph):
    ''' Render a vertically directed area between two equal length sequences
    of y-coordinates with the same x-coordinates using step lines.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    __example__ = "examples/reference/models/VAreaStep.py"

    _args = ('x', 'y1', 'y2')

    x = NumberSpec(default=field("x"), help="""
    The x-coordinates for the points of the area.
    """)

    y1 = NumberSpec(default=field("y1"), help="""
    The y-coordinates for the points of one side of the area.
    """)

    y2 = NumberSpec(default=field("y2"), help="""
    The y-coordinates for the points of the other side of the area.
    """)

    step_mode = Enum(StepMode, default="before", help="""
    Where the step "level" should be drawn in relation to the x and y
    coordinates. The parameter can assume one of three values:

    * ``before``: (default) Draw step levels before each x-coordinate (no step before the first point)
    * ``after``:  Draw step levels after each x-coordinate (no step after the last point)
    * ``center``: Draw step levels centered on each x-coordinate
    """)

    fill_props = Include(ScalarFillProps, help="""
    The {prop} values for the vertical directed area.
    """)

    hatch_props = Include(HatchProps, help="""
    The {prop} values for the vertical directed area.
    """)

class VBar(LRTBGlyph):
    ''' Render vertical bars, given a center coordinate, width and (top, bottom) coordinates.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    __example__ = "examples/reference/models/VBar.py"

    _args = ('x', 'width', 'top', 'bottom')

    x = NumberSpec(default=field("x"), help="""
    The x-coordinates of the centers of the vertical bars.
    """)

    width = DistanceSpec(default=1, help="""
    The widths of the vertical bars.
    """)

    bottom = NumberSpec(default=0, help="""
    The y-coordinates of the bottom edges.
    """)

    top = NumberSpec(default=field("top"), help="""
    The y-coordinates of the top edges.
    """)

    line_props = Include(LineProps, help="""
    The {prop} values for the vertical bars.
    """)

    fill_props = Include(FillProps, help="""
    The {prop} values for the vertical bars.
    """)

    hatch_props = Include(HatchProps, help="""
    The {prop} values for the vertical bars.
    """)

class Wedge(XYGlyph, LineGlyph, FillGlyph, HatchGlyph):
    ''' Render wedges.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    __example__ = "examples/reference/models/Wedge.py"

    _args = ('x', 'y', 'radius', 'start_angle', 'end_angle', 'direction')

    x = NumberSpec(default=field("x"), help="""
    The x-coordinates of the points of the wedges.
    """)

    y = NumberSpec(default=field("y"), help="""
    The y-coordinates of the points of the wedges.
    """)

    radius = DistanceSpec(default=field("radius"), help="""
    Radii of the wedges.
    """)

    start_angle = AngleSpec(default=field("start_angle"), help="""
    The angles to start the wedges, as measured from the horizontal.
    """)

    end_angle = AngleSpec(default=field("end_angle"), help="""
    The angles to end the wedges, as measured from the horizontal.
    """)

    direction = Enum(Direction, default='anticlock', help="""
    Which direction to stroke between the start and end angles.
    """)

    line_props = Include(LineProps, help="""
    The {prop} values for the wedges.
    """)

    fill_props = Include(FillProps, help="""
    The {prop} values for the wedges.
    """)

    hatch_props = Include(HatchProps, help="""
    The {prop} values for the wedges.
    """)

class HSpan(Glyph, LineGlyph):
    """ Horizontal lines of infinite width. """

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    __example__ = "examples/reference/models/HSpan.py"

    _args = ("y")

    y = NumberSpec(default=field("y"), help="""
    The y-coordinates of the spans.
    """)

    line_props = Include(LineProps, help="""
    The {prop} values for the spans.
    """)

class VSpan(Glyph, LineGlyph):
    """ Vertical lines of infinite height. """

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    __example__ = "examples/reference/models/VSpan.py"

    _args = ("x")

    x = NumberSpec(default=field("x"), help="""
    The x-coordinates of the spans.
    """)

    line_props = Include(LineProps, help="""
    The {prop} values for the spans.
    """)

class HStrip(Glyph, LineGlyph, FillGlyph, HatchGlyph):
    """ Horizontal strips of infinite width. """

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    __example__ = "examples/reference/models/HStrip.py"

    _args = ("y0", "y1")

    y0 = NumberSpec(default=field("y0"), help="""
    The y-coordinates of the coordinates of one side of the strips.
    """)

    y1 = NumberSpec(default=field("y1"), help="""
    The y-coordinates of the coordinates of the other side of the strips.
    """)

    line_props = Include(LineProps, help="""
    The {prop} values for the strips.
    """)

    fill_props = Include(FillProps, help="""
    The {prop} values for the strips.
    """)

    hatch_props = Include(HatchProps, help="""
    The {prop} values for the strips.
    """)

class VStrip(Glyph, LineGlyph, FillGlyph, HatchGlyph):
    """ Vertical strips of infinite height. """

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    __example__ = "examples/reference/models/VStrip.py"

    _args = ("x0", "x1")

    x0 = NumberSpec(default=field("x0"), help="""
    The x-coordinates of the coordinates of one side of the strips.
    """)

    x1 = NumberSpec(default=field("x1"), help="""
    The x-coordinates of the coordinates of the other side of the strips.
    """)

    line_props = Include(LineProps, help="""
    The {prop} values for the strips.
    """)

    fill_props = Include(FillProps, help="""
    The {prop} values for the strips.
    """)

    hatch_props = Include(HatchProps, help="""
    The {prop} values for the strips.
    """)

#-----------------------------------------------------------------------------
# Dev API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Private API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Code
#-----------------------------------------------------------------------------



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass
from inspect import Parameter
from typing import (
    Any,
    Final,
    Iterable,
    Self,
)

# Bokeh imports
from ..core.has_props import HasProps, Setter, abstract
from ..core.property.validation import without_property_validation
from ..core.query import SelectorType
from ..core.serialization import ObjectRefRep, Ref, Serializer
from ..core.types import ID
from ..document import Document
from ..document.events import DocumentPatchedEvent
from ..events import Event
from ..models.callbacks import (
    Callback as JSEventCallback,
    CustomCode as JSChangeCallback,
)
from ..util.callback_manager import (
    EventCallbackManager,
    PropertyCallback,
    PropertyCallbackManager,
)
from .util import HasDocumentRef

@abstract
@dataclass(init=False)
class Model(HasProps, HasDocumentRef, PropertyCallbackManager, EventCallbackManager):

    def destroy(self) -> None: ...

    id: Final[ID] = ...

    name: str | None = ...

    tags: list[Any] = ...

    js_event_callbacks: dict[str, list[JSEventCallback]] = ...

    js_property_callbacks: dict[str, list[JSEventCallback]] = ...

    subscribed_events: set[str] = ...

    syncable: bool = ...

    def __new__(cls, *args: Any, id: ID | None = None, **kwargs: Any) -> Self: ...

    @property
    def ref(self) -> Ref: ...

    @classmethod
    def clear_extensions(cls) -> None: ...

    @classmethod
    @without_property_validation
    def parameters(cls: type[Model]) -> list[Parameter]: ...

    def js_on_event(self, event: str | type[Event], *callbacks: JSEventCallback) -> None: ...

    def js_link(self, attr: str, other: Model, other_attr: str, attr_selector: int | str | None = None) -> None: ...

    def js_on_change(self, event: str, *callbacks: JSChangeCallback) -> None: ...

    def on_change(self, attr: str, *callbacks: PropertyCallback) -> None: ...

    def references(self) -> set[Model]: ...

    def select(self, selector: SelectorType) -> Iterable[Model]: ...

    def select_one(self, selector: SelectorType) -> Model | None: ...

    def set_select(self, selector: type[Model] | SelectorType, updates: dict[str, Any]) -> None: ...

    def to_serializable(self, serializer: Serializer) -> ObjectRefRep: ...

    def trigger(self, attr: str, old: Any, new: Any,
        hint: DocumentPatchedEvent | None = None, setter: Setter | None = None) -> None: ...

    def _attach_document(self, doc: Document) -> None: ...

    def _detach_document(self) -> None: ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass
from typing import Any

# Bokeh imports
from ..core.has_props import abstract
from ..model import Model

@abstract
@dataclass(init=False)
class Comparison(Model):
    ...

@dataclass
class CustomJSCompare(Comparison):

    args: dict[str, Any] = ...

    code: str = ...

@dataclass
class NanCompare(Comparison):

    ascending_first: bool = ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass
from typing import Any

# Bokeh imports
from ..model import Model

@dataclass
class TileSource(Model):

    url: str = ...

    tile_size: int = ...

    min_zoom: int = ...

    max_zoom: int = ...

    extra_url_vars: dict[str, Any] = ...

    attribution: str = ...

    x_origin_offset: float = ...

    y_origin_offset: float = ...

    initial_resolution: float | None = ...

@dataclass
class MercatorTileSource(TileSource):

    snap_to_zoom: bool = ...

    wrap_around: bool = ...

@dataclass
class TMSTileSource(MercatorTileSource):
    ...

@dataclass
class WMTSTileSource(MercatorTileSource):
    ...

@dataclass
class QUADKEYTileSource(MercatorTileSource):
    ...

@dataclass
class BBoxTileSource(MercatorTileSource):

    use_latlon: bool = ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass
from typing import Any, Sequence

# Bokeh imports
from ..core.enums import AutoType as Auto, LatLonType as LatLon
from ..core.has_props import abstract
from ..model import Model
from .mappers import ScanningColorMapper

@abstract
@dataclass(init=False)
class Ticker(Model):
    ...

@dataclass
class CustomJSTicker(Ticker):

    args: dict[str, Any] = ...

    major_code: str = ...

    minor_code: str = ...

@abstract
@dataclass(init=False)
class ContinuousTicker(Ticker):

    num_minor_ticks: int = ...

    desired_num_ticks: int = ...

@dataclass
class FixedTicker(ContinuousTicker):

    ticks: Sequence[float] = ...

    minor_ticks: Sequence[float] = ...

@dataclass
class AdaptiveTicker(ContinuousTicker):

    base: float = ...

    mantissas: Sequence[float] = ...

    min_interval: float = ...

    max_interval: float | None = ...

@dataclass
class CompositeTicker(ContinuousTicker):

    tickers: Sequence[Ticker] = ...

@dataclass
class BaseSingleIntervalTicker(ContinuousTicker):
    ...

@dataclass
class SingleIntervalTicker(BaseSingleIntervalTicker):

    interval: float = ...

@dataclass
class DaysTicker(BaseSingleIntervalTicker):

    days: Sequence[int] = ...

@dataclass
class MonthsTicker(BaseSingleIntervalTicker):

    months: Sequence[int] = ...

@dataclass
class YearsTicker(BaseSingleIntervalTicker):
    ...

@dataclass
class BasicTicker(AdaptiveTicker):
    ...

@dataclass
class LogTicker(AdaptiveTicker):
    ...

@dataclass
class MercatorTicker(BasicTicker):

    dimension: LatLon | None = ...

@dataclass
class CategoricalTicker(Ticker):
    ...

@dataclass
class DatetimeTicker(CompositeTicker):
    ...

@dataclass
class BinnedTicker(Ticker):

    mapper: ScanningColorMapper = ...

    num_major_ticks: int | Auto = ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import annotations # isort:skip

import pytest ; pytest

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

# Standard library imports
import os
import signal
import subprocess
import sys
import time
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Literal,
    NoReturn,
    TypeAlias,
)

if TYPE_CHECKING:
    from types import FrameType

# External imports
import _pytest.config
import _pytest.mark
import _pytest.python

#-----------------------------------------------------------------------------
# Globals and constants
#-----------------------------------------------------------------------------

BASE_DIR = Path(__file__).parent.parent
CASE_DIR = BASE_DIR / "tests" / "cross" / "cases"
BASELINE_DIR = BASE_DIR / "tests" / "baselines" / "cross"

#-----------------------------------------------------------------------------
# Setup
#-----------------------------------------------------------------------------

def pytest_generate_tests(metafunc: _pytest.python.Metafunc) -> None:
    if "case_path" not in metafunc.fixturenames:
        return
    config = metafunc.config
    params = [ pytest.param(str(path.relative_to(CASE_DIR)), config) for path in CASE_DIR.rglob("*.py") ]
    metafunc.parametrize("case_path,config", params)

#-----------------------------------------------------------------------------
# General API
#-----------------------------------------------------------------------------

@pytest.mark.skipif(sys.platform == "win32", reason="not supported on windows")
def test_cross(case_path: str, config: _pytest.config.Config) -> None:
    status, _, out, err = _run_test_case(CASE_DIR / case_path)
    if status == 0:
        baseline = (BASELINE_DIR / case_path).with_suffix(".json5")

        os.makedirs(baseline.parent, exist_ok=True)
        with baseline.open("w", encoding="utf-8") as file:
            file.write(out)

        baseline_path = baseline.relative_to(BASE_DIR)
        result = load_baseline(baseline_path)
        if result is None:
            assert False, "missing baseline; commit the baseline and re-run tests"

        status, out, _ = diff_baseline(baseline_path)
        if status != 0:
            print(out)
            assert False, "baseline differs"
    else:
        print(err)
        assert False, "test case failed to run"

#-----------------------------------------------------------------------------
# Dev API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Private API
#-----------------------------------------------------------------------------

ProcStatus: TypeAlias = int | Literal["timeout"]

TIMEOUT = 10 # seconds

def _run_test_case(path: Path) -> tuple[ProcStatus, float, str, str]:
    code = f"""\
__file__ = {str(path)!r}

import random
random.seed(1)

import numpy as np
np.random.seed(1)

import warnings
warnings.filterwarnings("ignore", ".*", UserWarning, "matplotlib.font_manager")

with open(__file__, "rb") as file:
    source = file.read()

global_vars = {{}}
exec(compile(source, __file__, "exec"), global_vars)

output = global_vars.get("output")
if output is None:
    raise RuntimeError("'output' not exported from a test case")

from bokeh.document import Document
from bokeh.model import Model

if isinstance(output, Model):
    doc = Document()
    doc.add_root(output)
elif isinstance(output, Document):
    doc = output
else:
    raise RuntimeError("invalid 'output', expected a 'Model' or 'Document'")

rep = doc.to_json(deferred=False)
del rep["title"]
del rep["version"]

import json5
json = json5.dumps(rep, sort_keys=False, indent=2)
print(json)
"""

    cmd = [sys.executable, "-c", code]
    cwd = path.parent

    env = os.environ.copy()

    class Timeout(Exception):
        pass

    if sys.platform != "win32":
        def alarm_handler(sig: int, frame: FrameType | None) -> NoReturn:
            raise Timeout

        signal.signal(signal.SIGALRM, alarm_handler)
        signal.alarm(TIMEOUT)

    start = time.time()
    with subprocess.Popen(cmd, cwd=cwd, env=env, stdout=subprocess.PIPE, stderr=subprocess.PIPE) as proc:
        status: ProcStatus
        try:
            status = proc.wait()
        except Timeout:
            proc.kill()
            status = "timeout"
        finally:
            if sys.platform != "win32":
                signal.alarm(0)

        end = time.time()

        assert proc.stdout is not None
        assert proc.stderr is not None

        out = proc.stdout.read().decode("utf-8")
        err = proc.stderr.read().decode("utf-8")

    return (status, end - start, out, err)

def git(*args: str) -> tuple[int, str, str]:
    proc = subprocess.Popen(["git", *args], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = proc.communicate()
    out = stdout.decode("utf-8", errors="ignore")
    err = stderr.decode("utf-8", errors="ignore")
    return (proc.returncode, out, err)

def load_baseline(baseline_path: Path, ref: str = "HEAD") -> str | None:
    status, out, _ = git("show", f"{ref}:./{baseline_path!s}")
    return out if status == 0 else None

def diff_baseline(baseline_path: Path, ref: str = "HEAD") -> tuple[int, str, str]:
    return git("diff", "--color", "--exit-code", ref, "--", str(baseline_path))

#-----------------------------------------------------------------------------
# Code
#-----------------------------------------------------------------------------



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass

# Bokeh imports
from ...core.has_props import abstract
from ..renderers.renderer import CompositeRenderer
from ..sources import DataSource

@abstract
@dataclass(init=False)
class Annotation(CompositeRenderer):
    ...

@abstract
@dataclass(init=False)
class DataAnnotation(Annotation):

    source: DataSource = ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------
""" Provide special versions of list and dict, that can automatically notify
about changes when used for property values.

Mutations to these values are detected, and the properties owning the
collection is notified of the changes. Consider the following model
definition:

.. code-block:: python

    class SomeModel(Model):

        options = List(String)

If we have an instance of this model, ``m`` then we can set the entire
value of the ``options`` property at once:

.. code-block:: python

    m.options = ["foo", "bar"]

When we do this in the context of a Bokeh server application that is being
viewed in a browser, this change is automatically noticed, and the
corresponding BokehJS property in the browser is synchronized, possibly
causing some change in the visual state of the application in the browser.

But it is also desirable that changes *inside* the ``options`` list also
be detected. That is, the following kinds of operations should also be
automatically synchronized between BokehJS and a Bokeh server:

.. code-block:: python

    m.options.append("baz")

    m.options[2] = "quux"

    m.options.insert(0, "bar")

The classes in this module provide this functionality.

.. note::
    These classes form part of the very low-level machinery that implements
    the Bokeh model and property system. It is unlikely that any of these
    classes or their methods will be applicable to any standard usage or to
    anyone who is not directly developing on Bokeh's own infrastructure.

"""

#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import annotations

import logging # isort:skip
log = logging.getLogger(__name__)

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

# Standard library imports
import copy
from typing import (
    TYPE_CHECKING,
    Any,
    Iterable,
    MutableSequence,
    Sequence,
    TypeVar,
)

# External imports
import numpy as np

if TYPE_CHECKING:
    import numpy.typing as npt

    from ...document import Document
    from ...document.events import DocumentPatchedEvent
    from ...models.sources import ColumnarDataSource
    from ..has_props import HasProps, Setter
    from .descriptors import PropertyDescriptor

#-----------------------------------------------------------------------------
# Globals and constants
#-----------------------------------------------------------------------------

__all__ = (
    'PropertyValueColumnData',
    'PropertyValueContainer',
    'PropertyValueDict',
    'PropertyValueList',
    'PropertyValueSet',
    'notify_owner',
)

T = TypeVar("T")

#----------------------------------------------------------------------------
# General API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Dev API
#-----------------------------------------------------------------------------

def notify_owner(func):
    """ A decorator for mutating methods of property container classes
    that notifies owners of the property container about mutating changes.

    Args:
        func (callable) : the container method to wrap in a notification

    Returns:
        wrapped method

    Examples:

        A ``__setitem__`` could be wrapped like this:

        .. code-block:: python

            # x[i] = y
            @notify_owner
            def __setitem__(self, i, y):
                return super().__setitem__(i, y)

    The returned wrapped method will have a docstring indicating what
    original method it is wrapping.

    """
    def wrapper(self, *args, **kwargs):
        old = self._saved_copy()
        result = func(self, *args, **kwargs)
        self._notify_owners(old)
        return result
    wrapper.__doc__ = f"Container method ``{func.__name__}`` instrumented to notify property owners"
    return wrapper

class PropertyValueContainer:
    """ A base class for property container classes that support change
    notifications on mutating operations.

    This class maintains an internal list of property owners, and also
    provides a private mechanism for methods wrapped with
    :func:`~bokeh.core.property.wrappers.notify_owners` to update
    those owners when mutating changes occur.

    """
    _owners: set[tuple[HasProps, PropertyDescriptor[Any]]]

    def __init__(self, *args, **kwargs) -> None:
        self._owners = set()
        super().__init__(*args, **kwargs)

    def _register_owner(self, owner: HasProps, descriptor: PropertyDescriptor[Any]) -> None:
        self._owners.add((owner, descriptor))

    def _unregister_owner(self, owner: HasProps, descriptor: PropertyDescriptor[Any]) -> None:
        self._owners.discard((owner, descriptor))

    def _notify_owners(self, old: Any, hint: DocumentPatchedEvent | None = None) -> None:
        for (owner, descriptor) in self._owners:
            descriptor._notify_mutated(owner, old, hint=hint)

    def _saved_copy(self) -> Any:
        raise RuntimeError("Subtypes must implement this to make a backup copy")

class PropertyValueList(PropertyValueContainer, list[T]):
    """ A list property value container that supports change notifications on
    mutating operations.

    When a Bokeh model has a ``List`` property, the ``PropertyValueLists`` are
    transparently created to wrap those values. These ``PropertyValueList``
    values are subject to normal property validation. If the property type
    ``foo = List(Str)`` then attempting to set ``x.foo[0] = 10`` will raise
    an error.

    Instances of ``PropertyValueList`` can be explicitly created by passing
    any object that the standard list initializer accepts, for example:

    .. code-block:: python

        >>> PropertyValueList([10, 20])
        [10, 20]

        >>> PropertyValueList((10, 20))
        [10, 20]

    The following mutating operations on lists automatically trigger
    notifications:

    .. code-block:: python

        del x[y]
        del x[i:j]
        x += y
        x *= y
        x[i] = y
        x[i:j] = y
        x.append
        x.extend
        x.insert
        x.pop
        x.remove
        x.reverse
        x.sort

    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def _saved_copy(self) -> list[T]:
        return list(self)

    # delete x[y]
    @notify_owner
    def __delitem__(self, y):
        return super().__delitem__(y)

    # x += y
    @notify_owner
    def __iadd__(self, y):
        return super().__iadd__(y)

    # x *= y
    @notify_owner
    def __imul__(self, y):
        return super().__imul__(y)

    # x[i] = y
    @notify_owner
    def __setitem__(self, i, y):
        return super().__setitem__(i, y)

    @notify_owner
    def append(self, obj):
        return super().append(obj)

    @notify_owner
    def extend(self, iterable):
        return super().extend(iterable)

    @notify_owner
    def insert(self, index, obj):
        return super().insert(index, obj)

    @notify_owner
    def pop(self, index=-1):
        return super().pop(index)

    @notify_owner
    def remove(self, obj):
        return super().remove(obj)

    @notify_owner
    def reverse(self):
        return super().reverse()

    @notify_owner
    def sort(self, **kwargs):
        return super().sort(**kwargs)

class PropertyValueSet(PropertyValueContainer, set[T]):
    """ A list property value container that supports change notifications on
    mutating operations.

    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def _saved_copy(self) -> set[T]:
        return set(self)

    @notify_owner
    def add(self, element: T) -> None:
        super().add(element)

    @notify_owner
    def difference_update(self, *s: Iterable[Any]) -> None:
        super().difference_update(*s)

    @notify_owner
    def discard(self, element: T) -> None:
        super().discard(element)

    @notify_owner
    def intersection_update(self, *s: Iterable[Any]) -> None:
        super().intersection_update(*s)

    @notify_owner
    def remove(self, element: T) -> None:
        super().discard(element)

    @notify_owner
    def symmetric_difference_update(self, s: Iterable[T]) -> None:
        super().symmetric_difference_update(s)

    @notify_owner
    def update(self, *s: Iterable[T]) -> None:
        super().update(*s)

T_Val = TypeVar("T_Val")

class PropertyValueDict(PropertyValueContainer, dict[str, T_Val]):
    """ A dict property value container that supports change notifications on
    mutating operations.

    When a Bokeh model has a ``List`` property, the ``PropertyValueLists`` are
    transparently created to wrap those values. These ``PropertyValueList``
    values are subject to normal property validation. If the property type
    ``foo = Dict(Str, Str)`` then attempting to set ``x.foo['bar'] = 10`` will
    raise an error.

    Instances of ``PropertyValueDict`` can be explicitly created by passing
    any object that the standard dict initializer accepts, for example:

    .. code-block:: python

        >>> PropertyValueDict(dict(a=10, b=20))
        {'a': 10, 'b': 20}

        >>> PropertyValueDict(a=10, b=20)
        {'a': 10, 'b': 20}

        >>> PropertyValueDict([('a', 10), ['b', 20]])
        {'a': 10, 'b': 20}

    The following mutating operations on dicts automatically trigger
    notifications:

    .. code-block:: python

        del x[y]
        x[i] = y
        x.clear
        x.pop
        x.popitem
        x.setdefault
        x.update

    """
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def _saved_copy(self):
        return dict(self)

    # delete x[y]
    @notify_owner
    def __delitem__(self, y):
        return super().__delitem__(y)

    # x[i] = y
    @notify_owner
    def __setitem__(self, i, y):
        return super().__setitem__(i, y)

    @notify_owner
    def clear(self):
        return super().clear()

    @notify_owner
    def pop(self, *args):
        return super().pop(*args)

    @notify_owner
    def popitem(self):
        return super().popitem()

    @notify_owner
    def setdefault(self, *args):
        return super().setdefault(*args)

    @notify_owner
    def update(self, *args, **kwargs):
        return super().update(*args, **kwargs)

class PropertyValueColumnData(PropertyValueDict[Sequence[Any]]):
    """ A property value container for ColumnData that supports change
    notifications on mutating operations.

    This property value container affords specialized code paths for
    updating the .data dictionary for ColumnDataSource. When possible,
    more efficient ColumnDataChangedEvent hints are generated to perform
    the updates:

    .. code-block:: python

        x[i] = y
        x.update

    """

    # x[i] = y
    # don't wrap with notify_owner --- notifies owners explicitly
    def __setitem__(self, i, y):
        return self.update([(i, y)])

    def __copy__(self):
        return PropertyValueColumnData(dict(self))

    def __deepcopy__(self, memodict={}):
        return PropertyValueColumnData(copy.deepcopy(dict(self), memodict))

    # don't wrap with notify_owner --- notifies owners explicitly
    def update(self, *args, **kwargs):
        old = self._saved_copy()

        # call dict.update directly, bypass wrapped version on base class
        result = dict.update(self, *args, **kwargs)

        from ...document.events import ColumnDataChangedEvent

        # Grab keys to update according to  Python docstring for update([E, ]**F)
        #
        # If E is present and has a .keys() method, then does:  for k in E: D[k] = E[k]
        # If E is present and lacks a .keys() method, then does:  for k, v in E: D[k] = v
        # In either case, this is followed by: for k in F:  D[k] = F[k]
        cols = set(kwargs.keys())
        if len(args) == 1:
            E = args[0]
            if hasattr(E, 'keys'):
                cols |= set(E.keys())
            else:
                cols |= { x[0] for x in E }

        # we must loop ourselves here instead of calling _notify_owners
        # because the hint is customized for each owner separately
        for (owner, descriptor) in self._owners:
            hint = ColumnDataChangedEvent(owner.document, owner, "data", cols=list(cols))
            descriptor._notify_mutated(owner, old, hint=hint)

        return result

    # don't wrap with notify_owner --- notifies owners explicitly
    def _stream(self, doc: Document, source: ColumnarDataSource, new_data: dict[str, Sequence[Any] | npt.NDArray[Any]],
            rollover: int | None = None, setter: Setter | None = None) -> None:
        """ Internal implementation to handle special-casing stream events
        on ``ColumnDataSource`` columns.

        Normally any changes to the ``.data`` dict attribute on a
        ``ColumnDataSource`` triggers a notification, causing all of the data
        to be synchronized between server and clients.

        The ``.stream`` method on column data sources exists to provide a
        more efficient way to perform streaming (i.e. append-only) updates
        to a data source, without having to perform a full synchronization,
        which would needlessly re-send all the data.

        To accomplish this, this function bypasses the wrapped methods on
        ``PropertyValueDict`` and uses the unwrapped versions on the dict
        superclass directly. It then explicitly makes a notification, adding
        a special ``ColumnsStreamedEvent`` hint to the message containing
        only the small streamed data that BokehJS needs in order to
        efficiently synchronize.

        .. warning::
            This function assumes the integrity of ``new_data`` has already
            been verified.

        """
        old = self._saved_copy()

        # TODO (bev) Currently this reports old differently for array vs list
        # For arrays is reports the actual old value. For lists, the old value
        # is actually the already updated value. This is because the method
        # self._saved_copy() makes a shallow copy.
        for k in new_data:
            old_seq = self[k]
            new_seq = new_data[k]

            if isinstance(old_seq, np.ndarray) or isinstance(new_seq, np.ndarray):
                # Special case for streaming with empty arrays, to allow this:
                #
                # data_source = ColumnDataSource(data={"DateTime": []})
                # data_source.stream({"DateTime": np.array([np.datetime64("now")]))
                #
                # See https://github.com/bokeh/bokeh/issues/14004.
                if len(old_seq) == 0:
                    seq = new_seq
                elif len(new_seq) == 0:
                    seq = old_seq
                else:
                    seq = np.append(old_seq, new_seq)

                if rollover is not None and len(seq) > rollover:
                    seq = seq[len(seq) - rollover:]

                # call dict.__setitem__ directly, bypass wrapped version on base class
                dict.__setitem__(self, k, seq)
            else:
                def apply_rollover(seq: MutableSequence[Any]) -> None:
                    if rollover is not None and len(seq) > rollover:
                        del seq[:len(seq) - rollover]

                if isinstance(old_seq, MutableSequence):
                    seq = old_seq
                    seq.extend(new_seq)
                    apply_rollover(seq)
                else:
                    seq = [*old_seq, *new_seq]
                    apply_rollover(seq)
                    dict.__setitem__(self, k, seq)

        from ...document.events import ColumnsStreamedEvent
        self._notify_owners(old, hint=ColumnsStreamedEvent(doc, source, "data", new_data, rollover, setter))

    # don't wrap with notify_owner --- notifies owners explicitly
    def _patch(self, doc: Document, source: ColumnarDataSource, patches, setter: Setter | None = None) -> None:
        """ Internal implementation to handle special-casing patch events
        on ``ColumnDataSource`` columns.

        Normally any changes to the ``.data`` dict attribute on a
        ``ColumnDataSource`` triggers a notification, causing all of the data
        to be synchronized between server and clients.

        The ``.patch`` method on column data sources exists to provide a
        more efficient way to perform patching (i.e. random access) updates
        to a data source, without having to perform a full synchronization,
        which would needlessly re-send all the data.

        To accomplish this, this function bypasses the wrapped methods on
        ``PropertyValueDict`` and uses the unwrapped versions on the dict
        superclass directly. It then explicitly makes a notification, adding
        a special ``ColumnsPatchedEvent`` hint to the message containing
        only the small patched data that BokehJS needs in order to efficiently
        synchronize.

        .. warning::
            This function assumes the integrity of ``patches`` has already
            been verified.

        """
        old = self._saved_copy()

        for name, patch in patches.items():
            for ind, value in patch:
                if isinstance(ind, int | slice):
                    self[name][ind] = value
                else:
                    shape = self[name][ind[0]][tuple(ind[1:])].shape
                    self[name][ind[0]][tuple(ind[1:])] = np.asarray(value).reshape(shape)

        from ...document.events import ColumnsPatchedEvent
        self._notify_owners(old, hint=ColumnsPatchedEvent(doc, source, "data", patches, setter))

#-----------------------------------------------------------------------------
# Private API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Code
#-----------------------------------------------------------------------------



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from contextlib import contextmanager
from dataclasses import dataclass
from typing import (
    Any,
    Generator,
    Sequence,
    TypeVar,
    overload,
)

# External imports
import xyzservices

# Bokeh imports
from .._types import Readonly
from ..core.enums import (
    LocationType as Location,
    OutputBackendType as OutputBackend,
    PlaceType as Place,
    ResetPolicyType as ResetPolicy,
    WindowAxisType as WindowAxis,
)
from ..core.property_aliases import LRTB
from ..core.property_mixins import (
    ScalarBackgroundFillProps as BackgroundFill,
    ScalarBackgroundHatchProps as BackgroundHatch,
    ScalarBorderFillProps as BorderFill,
    ScalarBorderHatchProps as BorderHatch,
    ScalarOutlineLineProps as OutlineLine,
)
from ..model import Model
from .annotations import Legend, Title
from .axes import Axis
from .dom import HTML
from .glyph import Glyph
from .grids import Grid
from .layouts import GridCommon, LayoutDOM
from .ranges import Range
from .renderers import GlyphRenderer, Renderer, TileRenderer
from .scales import Scale
from .sources import ColumnarDataSource
from .tiles import TileSource
from .tools import HoverTool, Tool, Toolbar

GlyphType = TypeVar("GlyphType", bound=Glyph)

class AxisListAttrSplat(list[Axis], Axis):
    pass

class GridListAttrSplat(list[Grid], Grid):
    pass

class LegendListAttrSplat(list[Legend], Legend):
    pass

class HoverListAttrSplat(list[HoverTool], HoverTool):
    pass

@dataclass
class Plot(LayoutDOM, BackgroundFill, BackgroundHatch, BorderFill, BorderHatch, OutlineLine):

    x_range: Range = ...

    y_range: Range = ...

    x_scale: Scale = ...

    y_scale: Scale = ...

    extra_x_ranges: dict[str, Range] = ...

    extra_y_ranges: dict[str, Range] = ...

    extra_x_scales: dict[str, Scale] = ...

    extra_y_scales: dict[str, Scale] = ...

    window_axis: WindowAxis = ...

    hidpi: bool = ...

    title: Title | str | None = ...

    title_location: Location | None = ...

    renderers: list[Renderer] = ...

    toolbar: Toolbar = ...

    toolbar_location: Location | None = ...

    toolbar_sticky: bool = ...

    toolbar_inner: bool = ...

    left: list[Renderer] = ...

    right: list[Renderer] = ...

    above: list[Renderer] = ...

    below: list[Renderer] = ...

    center: list[Renderer] = ...

    width: int | None = ...

    height: int | None = ...

    frame_width: int | None = ...

    frame_height: int | None = ...

    frame_align: bool | LRTB[bool] = ...

    inner_width: Readonly[int] = ...

    inner_height: Readonly[int] = ...

    outer_width: Readonly[int] = ...

    outer_height: Readonly[int] = ...

    min_border_top: int | None = ...

    min_border_bottom: int | None = ...

    min_border_left: int | None = ...

    min_border_right: int | None = ...

    min_border: int | None = ...

    lod_factor: int = ...

    lod_threshold: int | None = ...

    lod_interval: int = ...

    lod_timeout: int = ...

    output_backend: OutputBackend = ...

    match_aspect: bool = ...

    aspect_scale: float = ...

    reset_policy: ResetPolicy = ...

    hold_render: bool = ...

    attribution: list[HTML | str] = ...

    def select(self, *args: Any, **kwargs: Any) -> Sequence[Model]: ...

    def row(self, row: int, gridplot: GridPlot) -> bool: ...

    def column(self, col: int, gridplot: GridPlot) -> bool: ...

    @property
    def xaxis(self) -> AxisListAttrSplat: ...

    @property
    def yaxis(self) -> AxisListAttrSplat: ...

    @property
    def axis(self) -> AxisListAttrSplat: ...

    @property
    def legend(self) -> LegendListAttrSplat: ...

    @property
    def hover(self) -> HoverListAttrSplat: ...

    @property
    def xgrid(self) -> GridListAttrSplat: ...

    @property
    def ygrid(self) -> GridListAttrSplat: ...

    @property
    def grid(self) -> GridListAttrSplat: ...

    @property
    def tools(self) -> list[Tool]: ...

    @tools.setter
    def tools(self, tools: list[Tool]) -> None: ...

    def add_layout(self, obj: Renderer, place: Place = "center") -> None: ...

    def add_tools(self, *tools: Tool | str) -> None: ...

    def remove_tools(self, *tools: Tool) -> None: ...

    @overload
    def add_glyph(self, glyph: GlyphType, **kwargs: Any) -> GlyphRenderer[GlyphType]: ...
    @overload
    def add_glyph(self, source: ColumnarDataSource, glyph: GlyphType, **kwargs: Any) -> GlyphRenderer[GlyphType]: ...

    def add_tile(self, tile_source: TileSource | xyzservices.TileProvider | str, retina: bool = False, **kwargs: Any) -> TileRenderer: ...

    @contextmanager
    def hold(self, *, render: bool) -> Generator[None, None, None]: ...

@dataclass
class GridPlot(LayoutDOM, GridCommon):

    toolbar: Toolbar = ...

    toolbar_location: Location | None = ...

    children: list[tuple[LayoutDOM, int, int] | tuple[LayoutDOM, int, int, int, int]] = ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass

# Bokeh imports
from ...core.has_props import abstract
from ...model import Model

@abstract
@dataclass(init=False)
class Dimensional(Model):

    ticks: list[float] = ...

    include: list[str] | None = ...

    exclude: list[str] = ...

    def is_known(self, unit: str) -> bool: ...

@dataclass
class CustomDimensional(Dimensional):

    basis: dict[str, tuple[float, str] | tuple[float, str, str]] = ...

@dataclass
class Metric(Dimensional):

    base_unit: str = ...

    full_unit: str | None = ...

@dataclass
class ReciprocalMetric(Metric):
    ...

@dataclass
class MetricLength(Metric):
    ...

@dataclass
class ReciprocalMetricLength(ReciprocalMetric):
    ...

@dataclass
class ImperialLength(CustomDimensional):
    ...

@dataclass
class Angular(CustomDimensional):
    ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    Literal,
    Sequence,
    TypeAlias,
    overload,
)

# External imports
import numpy.typing as npt

if TYPE_CHECKING:
    import pandas as pd
    from pandas.core.groupby import GroupBy

# Bokeh imports
from .._types import JSON
from ..core.has_props import Setter, abstract
from ..model import Model
from .callbacks import CustomJS
from .filters import Filter
from .selections import Selection, SelectionPolicy

DataDict: TypeAlias = dict[str, Sequence[Any] | npt.NDArray[Any] | pd.Series[Any] | pd.Index[Any]]

DataDictLike: TypeAlias = DataDict | pd.DataFrame | GroupBy[Any]

Index: TypeAlias = int | slice | tuple[int | slice, ...]

Patches: TypeAlias = dict[str, list[tuple[Index, Any]]]

@abstract
@dataclass(init=False)
class DataSource(Model):

    selected: Selection = ...

@abstract
@dataclass(init=False)
class ColumnarDataSource(DataSource):

    default_values: dict[str, Any] = ...

    selection_policy: SelectionPolicy = ...

@dataclass
class ColumnDataSource(ColumnarDataSource):

    # TODO asymmetric get/set
    data: DataDictLike = ...

    @overload
    def __init__(self, data: DataDictLike, **kwargs: Any) -> None: ...
    @overload
    def __init__(self, **kwargs: Any) -> None: ...

    def __init__(self, *args: Any, **kwargs: Any) -> None: ...

    @property
    def column_names(self) -> list[str]: ...

    @classmethod
    def from_df(cls, data: pd.DataFrame) -> DataDict: ...

    @classmethod
    def from_groupby(cls, data: GroupBy[Any]) -> DataDict: ...

    def to_df(self) -> pd.DataFrame: ...

    def add(self, data: Sequence[Any], name: str | None) -> str: ...

    def remove(self, name: str) -> None: ...

    def stream(self, new_data: DataDict, rollover: int | None) -> None: ...

    def patch(self, patches: Patches, setter: Setter | None) -> None: ...

@dataclass
class CDSView(Model):

    filter: Filter = ...

@dataclass
class GeoJSONDataSource(ColumnarDataSource):

    geojson: JSON = ...

@abstract
@dataclass(init=False)
class WebDataSource(ColumnDataSource):

    adapter: CustomJS | None = ...

    max_size: int | None = ...

    mode: Literal["replace", "append"] = ...

    data_url: str = ...

@dataclass
class ServerSentDataSource(WebDataSource):
    ...

@dataclass
class AjaxDataSource(WebDataSource):

    polling_interval: int | None = ...

    method: Literal["POST", "GET"] = ...

    if_modified: bool = ...

    content_type: str = ...

    http_headers: dict[str, str] = ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass
from typing import NotRequired, TypedDict

# Bokeh imports
from .._types import NonNegative
from ..core.enums import (
    AlignType as Align,
    AutoType as Auto,
    DimensionsType as Dimensions,
    FlowModeType as FlowMode,
    LocationType as Location,
    ScrollbarPolicyType as ScrollbarPolicy,
    SizingModeType as SizingMode,
    SizingPolicyType as SizingPolicy,
)
from ..core.has_props import HasProps, abstract
from ..core.property_aliases import GridSpacing, TracksSizing
from ..model import Model
from .ui import Pane, Tooltip, UIElement

@abstract
@dataclass(init=False)
class LayoutDOM(Pane):

    disabled: bool = ...

    width: NonNegative[int] | None = ...

    height: NonNegative[int] | None = ...

    min_width: NonNegative[int] | None = ...

    min_height: NonNegative[int] | None = ...

    max_width: NonNegative[int] | None = ...

    max_height: NonNegative[int] | None = ...

    margin: int | tuple[int, int] | tuple[int, int, int, int] | None = ...

    width_policy: Auto | SizingPolicy = ...

    height_policy: Auto | SizingPolicy = ...

    aspect_ratio: None | Auto | float = ...

    flow_mode: FlowMode = ...

    sizing_mode: SizingMode | None = ...

    align: Auto | Align | tuple[Align, Align] = ...

    resizable: bool | Dimensions = ...

@dataclass
class Spacer(LayoutDOM):
    ...

@abstract
@dataclass(init=False)
class GridCommon(HasProps):

    rows: TracksSizing | None = ...

    cols: TracksSizing | None = ...

    spacing: GridSpacing = ...

@dataclass
class GridBox(LayoutDOM, GridCommon):

    children: list[tuple[UIElement, int, int] | tuple[UIElement, int, int, int, int]] = ...

class HBoxChild(TypedDict):
    child: UIElement
    col: NotRequired[int]
    span: NotRequired[int]

class VBoxChild(TypedDict):
    child: UIElement
    row: NotRequired[int]
    span: NotRequired[int]

@dataclass
class HBox(LayoutDOM):

    children: list[HBoxChild] | list[UIElement] = ...

    cols: TracksSizing | None = ...

    spacing: NonNegative[int] = ...

@dataclass
class VBox(LayoutDOM):

    children: list[VBoxChild] | list[UIElement] = ...

    rows: TracksSizing | None = ...

    spacing: NonNegative[int] = ...

@abstract
@dataclass(init=False)
class FlexBox(LayoutDOM):

    children: list[UIElement] = ...

    spacing: NonNegative[int] = ...

@dataclass
class Row(FlexBox):
    ...

@dataclass
class Column(FlexBox):
    ...

@dataclass
class TabPanel(Model):

    title: str = ...

    tooltip: Tooltip | None = ...

    child: UIElement = ...

    closable: bool = ...

    disabled: bool = ...

@dataclass
class Tabs(LayoutDOM):

    tabs: list[TabPanel] | list[tuple[str, UIElement]] = ...

    tabs_location: Location = ...

    active: int = ...

@dataclass
class GroupBox(LayoutDOM):

    title: str | None = ...

    child: UIElement = ...

    checkable: bool = ...

@dataclass
class ScrollBox(LayoutDOM):

    child: UIElement = ...

    horizontal_scrollbar: ScrollbarPolicy = ...

    vertical_scrollbar: ScrollbarPolicy = ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass

# Bokeh imports
from ...._specs import AngleSpec, NullStringSpec, NumberSpec
from ...._types import (
    Alpha,
    Angle,
    Color,
    CoordinateLike,
)
from ....core.enums import (
    AngleUnitsType as AngleUnits,
    CoordinateUnitsType as CoordinateUnits,
    FontStyleType as FontStyle,
    TextAlignType as TextAlign,
    VerticalAlignType as VerticalAlign,
)
from ....core.property_aliases import BorderRadius, Padding
from ....core.property_mixins import (
    BackgroundFillProps,
    BorderLineProps,
    ScalarBackgroundFillProps,
    ScalarBackgroundHatchProps,
    ScalarBorderLineProps,
    ScalarTextProps,
)
from ..annotation import DataAnnotation
from .html_annotation import HTMLAnnotation

@dataclass
class HTMLTextAnnotation(HTMLAnnotation, ScalarBackgroundFillProps, ScalarBackgroundHatchProps, ScalarBorderLineProps):

    padding: Padding = ...

    border_radius: BorderRadius = ...

@dataclass
class HTMLLabel(HTMLTextAnnotation, ScalarTextProps):

    x: CoordinateLike = ...

    x_units: CoordinateUnits = ...

    y: CoordinateLike = ...

    y_units: CoordinateUnits = ...

    text: str = ...

    angle: Angle = ...

    angle_units: AngleUnits = ...

    x_offset: float = ...

    y_offset: float = ...

@dataclass
class HTMLLabelSet(HTMLAnnotation, DataAnnotation, BackgroundFillProps, BorderLineProps):

    x: NumberSpec = ...

    x_units: CoordinateUnits = ...

    y: NumberSpec = ...

    y_units: CoordinateUnits = ...

    text: NullStringSpec = ...

    angle: AngleSpec = ...

    x_offset: NumberSpec = ...

    y_offset: NumberSpec = ...

@dataclass
class HTMLTitle(HTMLTextAnnotation):

    text: str = ...

    vertical_align: VerticalAlign = ...

    align: TextAlign = ...

    text_line_height: float = ...

    offset: float = ...

    standoff: float = ...

    text_font: str = ...

    text_font_size: str = ...

    text_font_style: FontStyle = ...

    text_color: Color = ...

    text_outline_color: Color | None = ...

    text_alpha: Alpha = ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass

# Bokeh imports
from ..core.enums import OutputBackendType as OutputBackend
from .ui import UIElement

@dataclass
class Canvas(UIElement):

    hidpi: bool = ...

    output_backend: OutputBackend = ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass

# Bokeh imports
from ..._specs import NumberSpec
from ...core.enums import CoordinateUnitsType as CoordinateUnits
from ...core.has_props import abstract
from ...core.property_mixins import (
    BodyLineProps,
    FillProps,
    HatchProps,
    LineProps,
)
from ..graphics import Marking
from .annotation import DataAnnotation

@abstract
@dataclass(init=False)
class ArrowHead(Marking):

    size: NumberSpec = ...

@dataclass
class OpenHead(ArrowHead, LineProps):
    ...

@dataclass
class NormalHead(ArrowHead, LineProps, FillProps, HatchProps):
    ...

@dataclass
class TeeHead(ArrowHead, LineProps):
    ...

@dataclass
class VeeHead(ArrowHead, LineProps, FillProps, HatchProps):
    ...

@dataclass
class Arrow(DataAnnotation, BodyLineProps):

    x_start: NumberSpec = ...

    y_start: NumberSpec = ...

    start_units: CoordinateUnits = ...

    start: ArrowHead | None = ...

    x_end: NumberSpec = ...

    y_end: NumberSpec = ...

    end_units: CoordinateUnits = ...

    end: ArrowHead | None = ...



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2023, Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------
''' Various kinds of indicator widgets.

'''

#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import annotations

import logging # isort:skip
log = logging.getLogger(__name__)

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

# Standard library imports
from typing import Any

# Bokeh imports
from ...core.enums import Orientation
from ...core.has_props import abstract
from ...core.properties import (
    Bool,
    Enum,
    Int,
    Nullable,
    String,
)
from .widget import Widget

#-----------------------------------------------------------------------------
# Globals and constants
#-----------------------------------------------------------------------------

__all__ = (
    "Progress",
)

#-----------------------------------------------------------------------------
# Dev API
#-----------------------------------------------------------------------------

@abstract
class Indicator(Widget):
    """ Abstract base class for indicator components.

    """

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

#-----------------------------------------------------------------------------
# General API
#-----------------------------------------------------------------------------

class Progress(Indicator):
    """ Linear progress indicator component.

    .. note::
        Long running or computationally heavy tasks may prevent the progress
        indicator from updating until such tasks are completed.
    """

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    mode = Enum("determinate", "indeterminate")(default="determinate", help="""
    Indicates the mode of operation of this widget.

    Supported modes are:
    * ``"determinate"`` where the component indicates progress of a task,
    computation, etc. until completion.
    * ``"indeterminate"`` where the component indicates stalled or infinitely
    running task. In this mode ``value``, ``min`` and ``max`` properties are
    ignored, and an infinite animation with no value label is displayed.

    .. note::
        The component can be switched between determinate and indeterminate
        modes during its operation.
    """)

    value = Int(default=0, help="""
    The current progress between the minimum and maximum values.
    """)

    min = Int(default=0, help="""
    The minimum value of the progress.
    """)

    max = Int(default=100, help="""
    The maximum value of the progress.
    """)

    reversed = Bool(default=False, help="""
    Allows to reverse progression of the progress.

    Depending on orientation this gives the following behaviors:

    * *left-to-right* when orientation is horizontal and not reversed
    * *right-to-left* when orientation is horizontal and reversed
    * *top-to-bottom* when orientation is vertical and not reversed
    * *bottom-to-top* when orientation is vertical and reversed
    """)

    orientation = Enum(Orientation, default="horizontal", help="""
    Allows to orient the progress bar either horizontally or vertically.
    """)

    label = Nullable(String)(default="@{percent}%", help="""
    Formatted value to show.

    May include the following special variables:

    * ``@{min}`` - the minimal value
    * ``@{max}`` - the maximal value
    * ``@{total}`` - the total number of steps
    * ``@{value}`` - the current value in the ``min`` to ``max`` range
    * ``@{index}`` - the current value in the 0 to ``total`` range
    * ``@{percent}`` - the current percentage of progress without ``%`` sign

    Special variables may additional specify a formatter, e.g. ``@{percent}{%.2f}``.
    """)

    label_location = Enum("none", "inline")(default="inline", help="""
    Determines where to place the label.
    """)

    description = Nullable(String)(help="""
    Optional long description to show in the from of a tooltip.

    This supports the same special variables as ``label`` property does.
    """)

#-----------------------------------------------------------------------------
# Private API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Code
#-----------------------------------------------------------------------------



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass

# Bokeh imports
from ...core.enums import (
    AnchorType as Anchor,
    AutoType as Auto,
    TooltipAttachmentType as TooltipAttachment,
)
from ..dom import DOMNode
from ..nodes import Coordinate
from ..selectors import Selector
from .ui_element import UIElement

@dataclass
class Tooltip(UIElement):

    position: Anchor | tuple[float, float] | Coordinate | None = ...

    target: UIElement | Selector | Auto = ...

    content: str | DOMNode | UIElement = ...

    attachment: TooltipAttachment | Auto = ...

    show_arrow: bool = ...

    closable: bool = ...

    interactive: bool = ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass

# Bokeh imports
from ...core.has_props import HasProps
from .ui_element import UIElement

@dataclass
class Examiner(UIElement):

    target: HasProps | None = ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------
''' Client-side interactivity.

'''

#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import annotations

import logging # isort:skip
log = logging.getLogger(__name__)

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

# Standard library imports
import pathlib
from typing import TYPE_CHECKING, Any

# Bokeh imports
from ..core.has_props import HasProps, abstract
from ..core.properties import (
    AnyRef,
    Auto,
    Bool,
    Dict,
    Either,
    Instance,
    Required,
    String,
)
from ..core.property.bases import Init
from ..core.property.singletons import Intrinsic
from ..core.validation import error
from ..core.validation.errors import INVALID_PROPERTY_VALUE, NOT_A_PROPERTY_OF
from ..model import Model

if TYPE_CHECKING:
    from ..core.types import PathLike

#-----------------------------------------------------------------------------
# Globals and constants
#-----------------------------------------------------------------------------

__all__ = (
    'Callback',
    'CloseDialog',
    'CustomJS',
    'OpenDialog',
    'OpenURL',
    'SetValue',
    'ToggleVisibility',
)

#-----------------------------------------------------------------------------
# General API
#-----------------------------------------------------------------------------

@abstract
class Callback(Model):
    ''' Base class for interactive callback.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

class OpenURL(Callback):
    ''' Open a URL in a new or current tab or window.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    url = String("http://", help="""
    The URL to direct the web browser to. This can be a template string,
    which will be formatted with data from the data source.
    """)

    same_tab = Bool(False, help="""
    Open URL in a new (`False`, default) or current (`True`) tab or window.
    For `same_tab=False`, whether tab or window will be opened is browser
    dependent.
    """)

@abstract
class CustomCode(Callback):
    """ """

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

class CustomJS(CustomCode):
    ''' Execute a JavaScript function.

    .. warning::
        The explicit purpose of this Bokeh Model is to embed *raw JavaScript
        code* for a browser to execute. If any part of the code is derived
        from untrusted user inputs, then you must take appropriate care to
        sanitize the user input prior to passing to Bokeh.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    args = Dict(String, AnyRef)(default={}, help="""
    A mapping of names to Python objects. In particular those can be bokeh's models.
    These objects are made available to the callback's code snippet as the values of
    named parameters to the callback.
    """)

    code = Required(String)(help="""
    A snippet of JavaScript code to execute in the browser.

    This can be interpreted either as a JavaScript function or a module, depending
    on the ``module`` property:

    1. A JS function.

    The code is made into the body of a function, and all of of the named objects in
    ``args`` are available as parameters that the code can use. Additionally,
    a ``cb_obj`` parameter contains the object that triggered the callback
    an optional ``cb_data`` parameter that contains any tool-specific data
    (i.e. mouse coordinates and hovered glyph indices for the ``HoverTool``)
    and additional document context in ``cb_context`` argument.

    2. An ES module.

    A JavaScript module (ESM) exporting a default function with the following
    signature:

    .. code-block: javascript

        export default function(args, obj, data, context) {
            // program logic
        }

    where ``args`` is a key-value mapping of user-provided parameters, ``obj``
    refers to the object that triggered the callback, ``data`` is a key-value
    mapping of optional parameters provided by the caller, and ``context`` is
    an additional document context.

    The additional document context is composed of the following members:

    * ``index``: The view manager governing all views in the current
      instance of ``Bokeh``. If only one instance of ``Bokeh`` is
      loaded, then this is equivalent to using ``Bokeh.index``.

    This function can be an asynchronous function (``async function () {}`` or
    ``async () => {}``) if for example external resources are needed, which
    would require usage of one of the asynchronous Web APIs, for example:

    .. code-block: javascript

        const response = await fetch("/assets/data.csv")
        const data = await response.text()

    """)

    module = Either(Auto, Bool, default="auto", help="""
    Whether to interpret the code as a JS function or ES module. If set to
    ``"auto"``, the this will be inferred from the code.
    """)

    @classmethod
    def from_file(cls, path: PathLike, **args: Any) -> CustomJS:
        """
        Construct a ``CustomJS`` instance from a ``*.js`` or ``*.mjs`` file.

        For example, if we want to construct a ``CustomJS`` instance from
        a JavaScript module ``my_module.mjs``, that takes a single argument
        ``source``, then we would use:

        .. code-block: python

            from bokeh.models import ColumnDataSource, CustomJS
            source = ColumnDataSource(data=dict(x=[1, 2, 3]))
            CustomJS.from_file("./my_module.mjs", source=source)

        """
        path = pathlib.Path(path)

        if path.suffix == ".js":
            module = False
        elif path.suffix == ".mjs":
            module = True
        else:
            raise RuntimeError(f"expected a *.js or *.mjs file, got {path}")

        with open(path, encoding="utf-8") as file:
            code = file.read()

        return CustomJS(code=code, args=args, module=module)

class SetValue(Callback):
    """ Allows to update a property of an object. """

    # explicit __init__ to support Init signatures
    def __init__(self, obj: Init[HasProps] = Intrinsic, attr: Init[str] = Intrinsic, value: Init[Any] = Intrinsic, **kwargs: Any) -> None:
        super().__init__(obj=obj, attr=attr, value=value, **kwargs)

    obj: HasProps = Required(Instance(HasProps), help="""
    Object to set the value on.
    """)

    attr: str = Required(String, help="""
    The property to modify.
    """)

    value = Required(AnyRef, help="""
    The value to set.
    """)

    @error(NOT_A_PROPERTY_OF)
    def _check_if_an_attribute_is_a_property_of_a_model(self):
        if self.obj.lookup(self.attr, raises=False):
            return None
        else:
            return f"{self.attr} is not a property of {self.obj}"

    @error(INVALID_PROPERTY_VALUE)
    def _check_if_provided_a_valid_value(self):
        descriptor = self.obj.lookup(self.attr)

        if descriptor.property.is_valid(self.value):
            return None
        else:
            return f"{self.value!r} is not a valid value for {self.obj}.{self.attr}"

class ToggleVisibility(Callback):
    """ Toggle visibility of a UI element. """

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    target = Required(Instance(".models.ui.UIElement"), help="""
    A UI element to show or hide.
    """)

class OpenDialog(Callback):
    """ Open a dialog box. """

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    dialog = Required(Instance(".models.ui.Dialog"), help="""
    A dialog instance to open.

    This will either build and display an new dialog view or re-open an
    existing dialog view if it was hidden.

    .. note::
        To display multiple instances of dialog, one needs to clone the
        dialog's model and use another instance of ``OpenDialog``.
    """)

class CloseDialog(Callback):
    """ Close a dialog box. """

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    dialog = Required(Instance(".models.ui.Dialog"), help="""
    A dialog instance to close.

    The behavior of this action depends on the configuration of the dialog,
    in particular ``Dialog.close_action`` property.
    """)

# TODO: class Show(Callback): target = Required(Either(Instance(DOMNode), Instance(UIElement)))
# TODO: class Hide(Callback): ...

#-----------------------------------------------------------------------------
# Dev API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Private API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Code
#-----------------------------------------------------------------------------



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass

# Bokeh imports
from ...core.enums import LocationType as Location
from ...core.property_aliases import CSSLength
from .panes import Pane

@dataclass
class Drawer(Pane):

    location: Location = ...

    open: bool = ...

    size: float | CSSLength = ...

    resizable: bool = ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------
""" Various kinds of icons to be used with Button widgets.
See :ref:`ug_interaction_widgets_examples_button` in the |user guide|
for more information.
"""

#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import annotations

import logging # isort:skip
log = logging.getLogger(__name__)

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

# Standard library imports
from typing import Any

# Bokeh imports
from ...core.enums import ToolIcon
from ...core.has_props import abstract
from ...core.properties import (
    Color,
    Either,
    Enum,
    FontSize,
    Int,
    Required,
    String,
)
from ...core.property.bases import Init
from ...core.property.singletons import Intrinsic
from .ui_element import UIElement

#-----------------------------------------------------------------------------
# Globals and constants
#-----------------------------------------------------------------------------

__all__ = (
    "Icon",
    "BuiltinIcon",
    "SVGIcon",
    "TablerIcon",
)

#-----------------------------------------------------------------------------
# General API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Dev API
#-----------------------------------------------------------------------------

@abstract
class Icon(UIElement):
    """ An abstract base class for icon elements.

    """

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    size = Either(Int, FontSize, default="1em", help="""
    The size of the icon. This can be either a number of pixels, or a CSS
    length string (see https://developer.mozilla.org/en-US/docs/Web/CSS/length).
    """)

class BuiltinIcon(Icon):
    """ Built-in icons included with BokehJS. """

    # explicit __init__ to support Init signatures
    def __init__(self, icon_name: Init[str] = Intrinsic, **kwargs: Any) -> None:
        super().__init__(icon_name=icon_name, **kwargs)

    icon_name = Required(Either(Enum(ToolIcon), String), help="""
    The name of a built-in icon to use. Currently, the following icon names are
    supported: ``"help"``, ``"question-mark"``, ``"settings"``, ``"x"``

    .. bokeh-plot::
        :source-position: none

        from bokeh.io import show
        from bokeh.layouts import column
        from bokeh.models import BuiltinIcon, Button

        builtin_icons = ["help", "question-mark", "settings", "x"]

        icon_demo = []
        for icon in builtin_icons:
            icon_demo.append(Button(label=icon, button_type="light", icon=BuiltinIcon(icon, size="1.2em")))

        show(column(icon_demo))

    """)

    color = Color(default="gray", help="""
    Color to use for the icon.
    """)

class SVGIcon(Icon):
    """ SVG icons with inline definitions. """

    # explicit __init__ to support Init signatures
    def __init__(self, svg: Init[str] = Intrinsic, **kwargs: Any) -> None:
        super().__init__(svg=svg, **kwargs)

    svg = Required(String, help="""
    The SVG definition of an icon.
    """)

class TablerIcon(Icon):
    """
    Icons from an external icon provider (https://tabler-icons.io/).

    .. note::
        This icon set is MIT licensed (see https://github.com/tabler/tabler-icons/blob/master/LICENSE).

    .. note::
        External icons are loaded from third-party servers and may not be available
        immediately (e.g. due to slow internet connection) or not available at all.
        It isn't possible to create a self-contained bundles with the use of
        ``inline`` resources. To circumvent this, one use ``SVGIcon``, by copying
        the SVG contents of an icon from Tabler's web site.

    """

    # explicit __init__ to support Init signatures
    def __init__(self, icon_name: Init[str] = Intrinsic, **kwargs: Any) -> None:
        super().__init__(icon_name=icon_name, **kwargs)

    icon_name = Required(String, help="""
    The name of the icon. See https://tabler-icons.io/ for the list of names.
    """)

#-----------------------------------------------------------------------------
# Private API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Code
#-----------------------------------------------------------------------------
